from ..utils import pickle_cached_container, pickle_cache_state
import pytest
import datetime
import os
import tempfile
import uuid


@pytest.fixture
def test_pickle_file():
    """Generate a unique temporary pickle file path for each test"""
    temp_file = os.path.join(tempfile.gettempdir(), f"test_cache_{uuid.uuid4().hex}.pickle")
    yield temp_file
    # Cleanup
    if os.path.exists(temp_file):
        os.unlink(temp_file)


def cache_file_exists(pickle_file):
    return os.path.isfile(pickle_file)


def kill_cache_file(pickle_file):
    if cache_file_exists(pickle_file):
        os.unlink(pickle_file)


def create_testing_class(pickle_file_name):
    """Factory function to create TestingClass with specific pickle file"""
    
    class TestingClass:
        def __init__(self):
            self.num_compute_calls = 0

        @pickle_cached_container()
        def my_tuple(self):
            return (1, 2, 3)

        @pickle_cached_container(pickle_file_name=pickle_file_name)
        def my_list(self):
            return self.compute_list()

        def compute_list(self):
            self.num_compute_calls += 1
            if not hasattr(self, "_return"):
                self.set_return([1, 2, 3])
            return self._return

        def set_return(self, val):
            self._return = val

        def get_num_calls(self):
            return self.num_compute_calls

        @classmethod
        def reset_call_count(cls):
            # This method is no longer needed but kept for compatibility
            pass
    
    return TestingClass




def test_out_of_scope(test_pickle_file):
    TestingClass = create_testing_class(test_pickle_file)
    try:
        kill_cache_file(test_pickle_file)
        TestingClass.reset_call_count()
        TestingClass.pkc = pickle_cache_state(mode="active")

        def scope(expected_calls):
            obj = TestingClass()
            assert obj.get_num_calls() == 0
            result = obj.my_list
            assert obj.get_num_calls() == expected_calls
            return result

        # Grabbing the results from an object once
        # will make it go out of scope.  I want to make sure
        # that the cachefile is not deleted when it goes out of scope
        scope(1)

        # Make sure the cache_file exists now that I'm out of scope
        assert cache_file_exists(test_pickle_file)

        # Calling again should not result in any additional compuations.
        scope(0)

    # Clean up
    finally:
        kill_cache_file(test_pickle_file)
        if hasattr(TestingClass, "pkc"):
            delattr(TestingClass, "pkc")


def check_pickle_cache_refresh_or_reset_mode(mode, test_pickle_file):
    TestingClass = create_testing_class(test_pickle_file)
    try:
        kill_cache_file(test_pickle_file)
        TestingClass.reset_call_count()

        # Set the testing class to active mode
        TestingClass.pkc = pickle_cache_state(mode=mode)

        obj = TestingClass()

        # Make sure no calls have been made an no cache file exists
        assert obj.get_num_calls() == 0
        assert not cache_file_exists(test_pickle_file)

        # This call should run a computation and file should exist
        result1 = obj.my_list
        assert obj.get_num_calls() == 1
        assert cache_file_exists(test_pickle_file)

        # Make sure the results looks right
        assert tuple(result1) == (1, 2, 3)

        # Now change the result that computation would return
        obj.set_return([4, 5, 6])

        # The result should now be different because another computation
        # should have been triggered
        result1 = obj.my_list
        assert obj.get_num_calls() == 2
        assert tuple(result1) == (4, 5, 6)
        assert cache_file_exists(test_pickle_file)

        # Creating a new object using active mode
        TestingClass.pkc = pickle_cache_state(mode="active")
        obj = TestingClass()

        # The new result should now be the new version from the pickle cache
        # and no computations should have been done
        result1 = obj.my_list
        assert obj.get_num_calls() == 0
        assert tuple(result1) == (4, 5, 6)

        # This is weird, but tuples return the same object when they are copied.
        # I guess it's because it doesn't make sense to copy immutable obj.
        assert id(obj.my_tuple) == id(obj.my_tuple)

    # Clean up
    finally:
        del obj.my_list
        del obj.my_tuple
        if hasattr(TestingClass, "pkc"):
            delattr(TestingClass, "pkc")


def test_bad_pickle_cache_mode():
    with pytest.raises(ValueError):
        pickle_cache_state(mode="this_is_bad")


def test_pickle_cache_reset_mode(test_pickle_file):
    return check_pickle_cache_refresh_or_reset_mode("reset", test_pickle_file)


def test_pickle_cache_refresh_mode(test_pickle_file):
    return check_pickle_cache_refresh_or_reset_mode("refresh", test_pickle_file)


def test_pickle_cache_memory_mode(test_pickle_file):
    # Create a separate pickle file for the active_obj test
    active_pickle_file = os.path.join(tempfile.gettempdir(), f"test_cache_active_{uuid.uuid4().hex}.pickle")
    
    TestingClass = create_testing_class(test_pickle_file)
    ActiveTestingClass = create_testing_class(active_pickle_file)
    
    try:
        kill_cache_file(test_pickle_file)
        TestingClass.reset_call_count()

        # Set the testing class to memory mode
        TestingClass.pkc = pickle_cache_state(mode="memory")

        obj = TestingClass()

        # Make sure no calls have been made an no cache file exists
        assert obj.get_num_calls() == 0
        assert not cache_file_exists(test_pickle_file)

        # This call should run a computation
        result1 = obj.my_list
        assert obj.get_num_calls() == 1

        # But no file should have been created
        assert not cache_file_exists(test_pickle_file)

        # Calling again should not trigger a computation or create a file
        result2 = obj.my_list
        assert obj.get_num_calls() == 1
        assert not cache_file_exists(test_pickle_file)

        # Make sure results look right
        assert tuple(result1) == (1, 2, 3)
        assert tuple(result2) == (1, 2, 3)

        # Now create another object that will create a cached file
        # With weird values
        ActiveTestingClass.pkc = pickle_cache_state(mode="active")
        active_obj = ActiveTestingClass()
        active_obj.set_return([7, 8, 9])
        active_obj.my_list
        assert cache_file_exists(active_pickle_file)
        TestingClass.pkc = pickle_cache_state(mode="memory")

        # Grab results from my original objects to make sure they don't
        # have the wacky values contained in the cache file
        result3 = obj.my_list
        assert obj.get_num_calls() == 1
        assert tuple(result3) == (1, 2, 3)

    # Clean up
    finally:
        del obj.my_list
        if hasattr(TestingClass, "pkc"):
            delattr(TestingClass, "pkc")
        # Clean up the active test file
        if os.path.exists(active_pickle_file):
            os.unlink(active_pickle_file)


# def test_pickle_cache_no_copy():
#     try:
#         kill_cache_file()

#         # Make sure the nocopy flag is actually returning th same
#         # object every time
#         obj = TestingClassNoCopy()
#         assert id(obj.my_list) == id(obj.my_list)

#     # Clean up
#     finally:
#         del obj.my_list
#         if hasattr(TestingClass, 'pkc'):
#             delattr(TestingClass, 'pkc')


def test_pickle_cache_active_mode(test_pickle_file):
    TestingClass = create_testing_class(test_pickle_file)
    try:
        kill_cache_file(test_pickle_file)
        TestingClass.reset_call_count()

        # Set the testing class to active mode
        TestingClass.pkc = pickle_cache_state(mode="active")

        obj = TestingClass()

        # Make sure no calls have been made an no cache file exists
        assert obj.get_num_calls() == 0
        assert not cache_file_exists(test_pickle_file)

        # This call should run a computation
        first_result = obj.my_list

        # Make sure the results look right
        assert tuple(first_result) == (1, 2, 3)

        # Make sure it was actually called and created a file
        assert obj.get_num_calls() == 1
        assert cache_file_exists(test_pickle_file)

        # This call should hit the memory cache
        second_result = obj.my_list

        # Make sure the results look right
        assert tuple(second_result) == (1, 2, 3)

        # Make sure the function was not evaluated
        assert obj.get_num_calls() == 1

        # Now delete the cache file and load property again
        kill_cache_file(test_pickle_file)
        third_result = obj.my_list

        # Make sure the result looks right
        assert tuple(third_result) == (1, 2, 3)

        # Function should still not be evaluated because data was in memory
        assert obj.get_num_calls() == 1

        # The call should not have recreated the cache file
        assert not cache_file_exists(test_pickle_file)

        # Now create a new object and make sure its value looks right
        obj = TestingClass()
        result4 = obj.my_list
        assert tuple(result4) == (1, 2, 3)

        # This should have called the compute function once
        assert obj.get_num_calls() == 1

        # And the cache file should now exist
        assert cache_file_exists(test_pickle_file)

        # Now create yes another object and make sure the results look right
        obj = TestingClass()
        result5 = obj.my_list
        assert tuple(result5) == (1, 2, 3)

        # Results should have been loaded from pickle file, so no calls
        assert obj.get_num_calls() == 0

    # Clean up
    finally:
        del obj.my_list
        if hasattr(TestingClass, "pkc"):
            delattr(TestingClass, "pkc")


def test_pickle_cache_ignore_mode(test_pickle_file):
    TestingClass = create_testing_class(test_pickle_file)
    try:
        TestingClass.reset_call_count()
        # Set the testing class to ignore mode
        TestingClass.pkc = pickle_cache_state(mode="ignore")

        # Call the property multiple times making sure it's computed each time
        obj = TestingClass()
        assert obj.get_num_calls() == 0
        obj.my_list
        assert obj.get_num_calls() == 1
        obj.my_list
        assert obj.get_num_calls() == 2

    # Clean up
    finally:
        del obj.my_list
        if hasattr(TestingClass, "pkc"):
            delattr(TestingClass, "pkc")


def test_pickle_cache_default_file_creation():
    try:
        today = str(datetime.datetime.now().date())
        # Use unique name to avoid conflicts between parallel tests
        unique_id = uuid.uuid4().hex[:8]
        expected_name = "/tmp/TestingClassForFileCreation_{}.my_list_{}.pickle".format(
            unique_id, today
        )

        # Delete any existing picklefile and make sure it is gone
        if os.path.isfile(expected_name):
            os.unlink(expected_name)
        assert not os.path.isfile(expected_name)

        # Create a dynamic class with unique pickle file name
        class TestingClassForFileCreationUnique:
            @pickle_cached_container(pickle_file_name=expected_name)
            def my_list(self):
                return [1, 2, 3]

        # Make sure the property has the expected value
        obj = TestingClassForFileCreationUnique()
        assert obj.my_list == [1, 2, 3]

        # Make sure the expected cache file exists
        assert os.path.isfile(expected_name)

        # Make sure each call to the cache returns a new identical object
        assert obj.my_list == obj.my_list

        # import pdb; pdb.set_trace()
        assert id(obj.my_list) != id(obj.my_list)

    finally:
        del obj.my_list
        # Clean up the unique file
        if os.path.exists(expected_name):
            os.unlink(expected_name)


def test_pickle_cache_custom_file_creation():
    try:
        # Use unique name to avoid conflicts between parallel tests
        unique_id = uuid.uuid4().hex[:8]
        expected_name = f"/tmp/silly_pickle_{unique_id}.pickle"

        # Delete any existing picklefile and make sure it is gone
        if os.path.isfile(expected_name):
            os.unlink(expected_name)
        assert not os.path.isfile(expected_name)

        # Create a dynamic class with unique pickle file name
        class TestingClassForFileCreationCustomUnique:
            @pickle_cached_container(
                pickle_file_name=expected_name, return_copy=False
            )
            def my_list(self):
                return [1, 2, 3]

        # Make sure the property has the expected value
        obj = TestingClassForFileCreationCustomUnique()
        assert obj.my_list == [1, 2, 3]

        # Make sure the expected cache file exists
        assert os.path.isfile(expected_name)

        # Make sure each call to the cache returns a new identical object
        assert obj.my_list == obj.my_list

        # Make sure the same object is returned
        assert id(obj.my_list) == id(obj.my_list)

    finally:
        del obj.my_list
        # Clean up the unique file
        if os.path.exists(expected_name):
            os.unlink(expected_name)



import pandas as pd
from easier.duckcacher import DuckCacher, duckloader_factory
from unittest.mock import patch
import tempfile
import os
import shutil
import pytest
from easier.utils import cached_container
import polars as pl
import duckdb

################ To run tests ######################################
# python -m pytest -v --cov=duckcacher --cov-report=term-missing
######################################################################


def generate_test_df(df_type: str) -> pd.DataFrame:
    """Generate a test dataframe based on the specified type.

    Args:
        df_type (str): Type of dataframe to generate. Options:
            - 'numbers': Simple numeric dataframe
            - 'text': Simple text dataframe
            - 'mixed': Mixed numeric and text dataframe
            - 'empty': Empty dataframe with columns
            - 'dates': Dataframe with dates

    Returns:
        pd.DataFrame: Generated test dataframe
    """
    if df_type == "numbers":
        return pd.DataFrame({"id": [1, 2, 3], "value": [10.5, 20.5, 30.5]})
    elif df_type == "text":
        return pd.DataFrame(
            {"name": ["Alice", "Bob", "Charlie"], "category": ["A", "B", "A"]}
        )
    elif df_type == "mixed":
        return pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "score": [95.5, 88.0, 92.5],
            }
        )
    elif df_type == "empty":
        return pd.DataFrame(columns=["id", "name", "value"])
    elif df_type == "dates":
        return pd.DataFrame(
            {"date": pd.date_range("2024-01-01", periods=3), "value": [100, 200, 300]}
        )
    else:
        raise ValueError(f"Unknown dataframe type: {df_type}")


@pytest.fixture
def temp_db():
    """Create a temporary database file for testing."""
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "test.db")
    yield db_path
    shutil.rmtree(temp_dir)


def test_register_decorator(temp_db):
    """Test that the register decorator correctly stores and returns dataframes."""
    # Create a new DuckCacher instance with temporary database
    cache = DuckCacher(temp_db)

    # Create a test function that returns a mixed dataframe
    @cache.register
    def test_table() -> pd.DataFrame:
        return generate_test_df("mixed")

    # Get the dataframe through the cache
    result_df = cache.df_test_table

    # Verify the dataframe contents
    expected_df = generate_test_df("mixed")
    pd.testing.assert_frame_equal(result_df, expected_df)

    # Verify we can access the same dataframe multiple times
    result_df2 = cache.df_test_table
    pd.testing.assert_frame_equal(result_df2, expected_df)


def test_register_as_basic(temp_db):
    """Test that register_as correctly stores and returns dataframes with custom table names."""
    cache = DuckCacher(temp_db)

    @cache.register_as("custom_table")
    def some_function() -> pd.DataFrame:
        return generate_test_df("mixed")

    # Get the dataframe through the cache using the custom name
    result_df = cache.df_custom_table

    # Verify the dataframe contents
    expected_df = generate_test_df("mixed")
    pd.testing.assert_frame_equal(result_df, expected_df)

    # Verify we can access the same dataframe multiple times
    result_df2 = cache.df_custom_table
    pd.testing.assert_frame_equal(result_df2, expected_df)


def test_register_as_function_name_independence(temp_db):
    """Test that the function name doesn't affect the table name when using register_as."""
    cache = DuckCacher(temp_db)

    @cache.register_as("my_table")
    def completely_different_name() -> pd.DataFrame:
        return generate_test_df("numbers")

    # Access should work with the custom table name, not the function name
    result_df = cache.df_my_table
    expected_df = generate_test_df("numbers")
    pd.testing.assert_frame_equal(result_df, expected_df)

    # Verify we can't access using the function name
    with pytest.raises(AttributeError):
        _ = cache.df_completely_different_name


def test_register_as_caching(temp_db):
    """Test that register_as has the same caching behavior as register."""
    cache = DuckCacher(temp_db)
    call_count = 0

    @cache.register_as("cached_table")
    def some_function() -> pd.DataFrame:
        nonlocal call_count
        call_count += 1
        return generate_test_df("numbers")

    # First call should execute the function
    df1 = cache.df_cached_table
    assert call_count == 1, "Function should be called on first access"

    # Second call should use cache
    df2 = cache.df_cached_table
    assert call_count == 1, "Function should not be called again"

    # Verify we get copies of the dataframe
    assert df1 is not df2, "Should get different dataframe objects"
    pd.testing.assert_frame_equal(df1, df2)


def test_register_as_duplicate_table(temp_db):
    """Test that we can't register two functions with the same table name."""
    cache = DuckCacher(temp_db)

    @cache.register_as("duplicate_table")
    def first_function() -> pd.DataFrame:
        return generate_test_df("numbers")

    # Registering another function with the same table name should raise an error
    with pytest.raises(ValueError):

        @cache.register_as("duplicate_table")
        def second_function() -> pd.DataFrame:
            return generate_test_df("text")


def test_caching_behavior(temp_db):
    """Test that the caching mechanism works correctly and returns copies."""
    # Create a new DuckCacher instance with temporary database
    cache = DuckCacher(temp_db)

    # Track function calls
    call_count = 0

    @cache.register
    def cached_table() -> pd.DataFrame:
        nonlocal call_count
        call_count += 1
        return generate_test_df("numbers")

    # First call should execute the function
    df1 = cache.df_cached_table
    assert call_count == 1, "Function should be called on first access"

    # Second call should use cache
    df2 = cache.df_cached_table
    assert call_count == 1, "Function should not be called again"

    # Verify we get copies of the dataframe
    assert df1 is not df2, "Should get different dataframe objects"

    # Verify the dataframes are equal
    pd.testing.assert_frame_equal(df1, df2)

    # Verify modifying one dataframe doesn't affect the other
    df1.loc[0, "value"] = 999.9
    assert (
        df1.loc[0, "value"] != df2.loc[0, "value"]
    ), "Modifying one dataframe should not affect the other"


def test_cache_contents(temp_db):
    """Test that the _cache attribute contains a copy of the dataframe only when cache_in_memory=True."""
    # Test with cache_in_memory=True
    cache = DuckCacher(temp_db, cache_in_memory=True)

    @cache.register
    def test_table() -> pd.DataFrame:
        return generate_test_df("numbers")

    # First access should populate cache
    df1 = cache.df_test_table
    expected_df = generate_test_df("numbers")

    # Verify cache contains a copy of the dataframe
    cache_key = "main.test_table"
    assert cache_key in cache._cache, "Cache should contain the dataframe"
    pd.testing.assert_frame_equal(cache._cache[cache_key], expected_df)
    assert (
        cache._cache[cache_key] is not df1
    ), "Cache should contain a copy, not the same object"

    # Test with cache_in_memory=False
    cache = DuckCacher(temp_db, cache_in_memory=False)

    @cache.register
    def test_table2() -> pd.DataFrame:
        return generate_test_df("numbers")

    # Access should not populate cache
    df2 = cache.df_test_table2

    # Verify cache is empty
    assert len(cache._cache) == 0, "Cache should be empty when cache_in_memory=False"


def test_run_query_call_count(temp_db):
    # cache_in_memory=True: run_query should be called only once for multiple accesses
    cache = DuckCacher(temp_db, cache_in_memory=True)
    with patch.object(
        cache, "run_query_df", wraps=cache.run_query_df
    ) as mock_run_query:

        @cache.register
        def test_table() -> pd.DataFrame:
            return generate_test_df("numbers")

        # First access triggers run_query for CREATE TABLE and SELECT
        df1 = cache.df_test_table
        expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df1, expected_df)

        # Verify no queries have been called yet
        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

        df2 = cache.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)
        assert df1 is not df2, "Dataframes should be different objects (copies)"
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

    cache = DuckCacher(temp_db, cache_in_memory=False)
    with patch.object(
        cache, "run_query_df", wraps=cache.run_query_df
    ) as mock_run_query:

        @cache.register
        def test_table() -> pd.DataFrame:
            return generate_test_df("numbers")

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

        # First access triggers run_query for CREATE TABLE and SELECT
        df1 = cache.df_test_table
        expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df1, expected_df)

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 1
        ), "One SELECT query should have been executed for test_table"

        df2 = cache.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)
        assert df1 is not df2, "Dataframes should be different objects (copies)"

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 2
        ), "Two SELECT query should have been executed for test_table"


def test_file_based_database():
    """Test DuckCacher functionality with a file-based database."""
    # Create a temporary directory for our test database
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "test.db")

    try:
        # Cached file should not exist
        assert not os.path.exists(db_path), "Database file should not exist"

        # Create a new DuckCacher instance with file-based database
        cache = DuckCacher(db_path, cache_in_memory=False)

        # Track function calls
        call_count = 0

        # Register a test function
        @cache.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count
            call_count += 1
            return generate_test_df("mixed")

        # Verify the database file exists
        assert os.path.exists(db_path), "Database file should be created"

        # First access should return expected data
        df1 = cache.df_test_table
        expected_df = generate_test_df("mixed")
        pd.testing.assert_frame_equal(df1, expected_df)

        # Function should be called once on first access
        assert call_count == 1, "Function should be called once on first access"

        _ = cache.df_test_table

        # Function should not be called on second call
        assert call_count == 1, "Function should be called only once"

        # Create a new cache instance pointing to the same database
        cache2 = DuckCacher(db_path, cache_in_memory=False)

        # Reset call count for second cache
        call_count = 0

        # Register the same function
        @cache2.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count
            call_count += 1
            return generate_test_df("mixed")

        # Access should return the same data
        df2 = cache2.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)

        # Second cache should not have called the function
        assert call_count == 0, "Function should not be called on second cache"

        # Modify the function to return different data
        call_count = 0  # Reset counter for modified function

        @cache.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count
            call_count += 1
            return generate_test_df("numbers")

        # Access should return the new data
        df3 = cache.df_test_table
        new_expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df3, new_expected_df)
        assert (
            call_count == 1
        ), "Modified function should be called once on first access"

        # Second access should use cache
        df3_2 = cache.df_test_table
        pd.testing.assert_frame_equal(df3_2, new_expected_df)
        assert (
            call_count == 1
        ), "Modified function should not be called again on second access"

        # Verify we get copies of the dataframes
        assert df1 is not df2, "Dataframes should be different objects"
        assert df2 is not df3, "Dataframes should be different objects"
        assert df3 is not df3_2, "Dataframes should be different objects (copies)"

        # Verify the dataframes are not equal after modification
        pd.testing.assert_frame_equal(df1, df2)  # Should be equal (both mixed)
        with pytest.raises(AssertionError):
            pd.testing.assert_frame_equal(
                df1, df3
            )  # Should not be equal (mixed vs numbers)

    finally:
        # Clean up the temporary directory
        shutil.rmtree(temp_dir)


def test_overwrite_removes_existing_file():
    """Test that the overwrite parameter correctly handles existing database files."""
    # Create a temporary directory for our test
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "overwrite_test.db")

    try:
        # Phase 1: Initial database creation
        assert not os.path.exists(db_path), "Database file should not exist initially"

        cache = DuckCacher(db_path, overwrite=False, cache_in_memory=False)

        @cache.register
        def test_table() -> pd.DataFrame:
            return pd.DataFrame({"a": [1, 2, 3]})

        df1 = cache.df_test_table
        pd.testing.assert_frame_equal(df1, pd.DataFrame({"a": [1, 2, 3]}))
        assert os.path.exists(
            db_path
        ), "Database file should exist after first creation"

        # Get initial file creation time
        initial_file_stats = os.stat(db_path)
        initial_creation_time = initial_file_stats.st_ctime

        # Phase 2: Overwrite with new data
        # File should still exist before overwrite
        assert os.path.exists(db_path), "Database file should exist before overwrite"

        # Create a new cache with overwrite=True - this should delete the existing file
        cache2 = DuckCacher(db_path, overwrite=True, cache_in_memory=False)

        # Verify the file was deleted and recreated by checking creation time
        assert os.path.exists(db_path), "Database file should exist after overwrite"
        new_file_stats = os.stat(db_path)
        new_creation_time = new_file_stats.st_ctime
        assert (
            new_creation_time > initial_creation_time
        ), "File should have been deleted and recreated with a newer timestamp"

        # Register and access the table to ensure it's using the new database
        @cache2.register
        def test_table() -> pd.DataFrame:
            return pd.DataFrame({"a": [42, 43, 44]})

        df2 = cache2.df_test_table
        pd.testing.assert_frame_equal(df2, pd.DataFrame({"a": [42, 43, 44]}))

        # Phase 3: Verify data was overwritten
        assert os.path.exists(
            db_path
        ), "Database file should still exist after overwrite"
        assert not df2.equals(
            pd.DataFrame({"a": [1, 2, 3]})
        ), "Old data should not be present after overwrite"

    finally:
        # Clean up
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)


def test_getattr_errors(temp_db):
    """Test that __getattr__ raises appropriate errors for invalid attributes."""
    cache = DuckCacher(temp_db)
    with pytest.raises(AttributeError, match="has no attribute 'invalid'"):
        _ = cache.invalid
    with pytest.raises(AttributeError, match="has no attribute 'df_unregistered'"):
        _ = cache.df_unregistered


def test_multiple_cacher_instances_with_mirrored_objects(temp_db):
    """Test caching behavior when multiple DuckCacher instances point to the same database file
    and mirror different instances of the same class."""

    # Create a class with cached_container attributes
    class TestData:
        def __init__(self, df_type):
            self.df_type = df_type

        @cached_container
        def df_numbers(self):
            return generate_test_df("numbers")

        @cached_container
        def df_text(self):
            return generate_test_df("text")

        @cached_container
        def df_mixed(self):
            return generate_test_df("mixed")

    # Create two instances of the class
    data1 = TestData("numbers")
    data2 = TestData("text")

    # Create two DuckCacher instances pointing to the same database file
    cache1 = DuckCacher(temp_db, mirrored_object=data1)
    cache2 = DuckCacher(temp_db, mirrored_object=data2)

    # Access the dataframes through both caches
    df1_numbers = cache1.df_numbers
    df2_numbers = cache2.df_numbers

    # Verify the dataframes are equal since they use the same generator
    pd.testing.assert_frame_equal(df1_numbers, df2_numbers)

    # Verify the text dataframe is the same for both instances
    df1_text = cache1.df_text
    df2_text = cache2.df_text
    pd.testing.assert_frame_equal(df1_text, df2_text)

    # Verify caching works by accessing again
    df1_numbers_2 = cache1.df_numbers
    df2_numbers_2 = cache2.df_numbers

    # Should get copies of the dataframes
    assert df1_numbers is not df1_numbers_2
    assert df2_numbers is not df2_numbers_2

    # But the data should be the same
    pd.testing.assert_frame_equal(df1_numbers, df1_numbers_2)
    pd.testing.assert_frame_equal(df2_numbers, df2_numbers_2)


def test_duck_loader():
    """Test that DuckLoader correctly creates cached attributes for database tables."""
    # Create a temporary directory for our test database
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "test.db")

    try:
        # Create a new DuckCacher instance to set up some test tables
        cache = DuckCacher(db_path)

        # Create test tables with and without df_ prefix
        @cache.register
        def test_table1() -> pd.DataFrame:
            return generate_test_df("numbers")

        @cache.register
        def test_table2() -> pd.DataFrame:
            return generate_test_df("text")

        @cache.register
        def df_test_table3() -> pd.DataFrame:
            return generate_test_df("text")

        # Access the tables to ensure they're created
        _ = cache.df_test_table1
        _ = cache.df_test_table2
        _ = cache.df_df_test_table3

        # Create a DuckLoader instance using the factory function
        loader = duckloader_factory(db_path)

        # Test ls() method
        loaded_tables = loader.ls()
        assert len(loaded_tables) == 3, "Should have 3 loaded tables"
        assert (
            "df_test_table1" in loaded_tables
        ), "Should include test_table1 with df_ prefix"
        assert (
            "df_test_table2" in loaded_tables
        ), "Should include test_table2 with df_ prefix"
        assert "df_test_table3" in loaded_tables, "Should include df_test_table3 as is"

        # Test accessing tables with and without df_ prefix
        df1 = loader.df_test_table1  # Should add df_ prefix
        df2 = loader.df_test_table2  # Should add df_ prefix
        df3 = loader.df_test_table3  # Should keep original name with df_ prefix

        # Verify the data is correct
        pd.testing.assert_frame_equal(df1, generate_test_df("numbers"))
        pd.testing.assert_frame_equal(df2, generate_test_df("text"))
        pd.testing.assert_frame_equal(df3, generate_test_df("text"))

        # Verify caching works by accessing again
        df1_2 = loader.df_test_table1
        df2_2 = loader.df_test_table2
        df3_2 = loader.df_test_table3

        # Verify we get copies
        assert df1 is not df1_2, "Should get a copy of the dataframe"
        assert df2 is not df2_2, "Should get a copy of the dataframe"
        assert df3 is not df3_2, "Should get a copy of the dataframe"

        # Verify the data is still correct
        pd.testing.assert_frame_equal(df1_2, generate_test_df("numbers"))
        pd.testing.assert_frame_equal(df2_2, generate_test_df("text"))
        pd.testing.assert_frame_equal(df3_2, generate_test_df("text"))

        # Test accessing with original table names (should fail)
        with pytest.raises(AttributeError):
            _ = loader.test_table1
        with pytest.raises(AttributeError):
            _ = loader._private_table

        # Test accessing non-existent table
        with pytest.raises(AttributeError, match="has no attribute 'df_nonexistent'"):
            _ = loader.df_nonexistent

    finally:
        # Clean up
        shutil.rmtree(temp_dir)


def test_in_memory_database_error():
    """Test that using ':memory:' as file_name raises an error."""
    with pytest.raises(ValueError, match="In-memory databases are not supported"):
        DuckCacher(":memory:")


def test_get_cached_dataframe_error(temp_db):
    """Test error handling in _get_cached_dataframe."""
    cache = DuckCacher(temp_db)

    # Create a class that will raise AttributeError on clone/copy
    class BadDataFrame(pd.DataFrame):
        def clone(self):
            raise AttributeError("Bad attribute error")

        def copy(self, deep=True):  # Fixed: Added deep parameter
            raise AttributeError("Bad attribute error")

    # Register a function that returns our problematic DataFrame
    @cache.register
    def bad_table() -> pd.DataFrame:
        return BadDataFrame({"col": [1, 2, 3]})

    # Accessing the table should raise AttributeError
    with pytest.raises(AttributeError, match="Bad attribute error"):
        _ = cache.df_bad_table


def test_needs_sync_error_handling(temp_db):
    """Test error handling in _needs_sync."""
    cache = DuckCacher(temp_db)

    # Test with empty DataFrame but with columns
    @cache.register
    def empty_table() -> pd.DataFrame:
        return pd.DataFrame(columns=["col1", "col2"])

    # Force a sync to create the table
    cache.sync("empty_table")

    # The function should handle empty DataFrames gracefully
    assert cache._needs_sync("main", "empty_table", "some_hash")


def test_sync_error_handling(temp_db):
    """Test error handling in sync method."""
    cache = DuckCacher(temp_db)

    # Test with non-existent table
    with pytest.raises(ValueError, match="No registered function found for table"):
        cache.sync("non_existent_table")

    # Test with invalid table name
    with pytest.raises(ValueError, match="No registered function found for table"):
        cache.sync("invalid/table/name")


def test_register_as_error_handling(temp_db):
    """Test error handling in register_as decorator."""
    cache = DuckCacher(temp_db)

    # Test with invalid table name
    with pytest.raises(
        ValueError,
        match=(
            "Invalid table name 'invalid/name'. Table names must start with a letter or underscore "
            "and contain only alphanumeric characters and underscores."
        ),
    ):

        @cache.register_as("invalid/name")
        def invalid_name_func() -> pd.DataFrame:
            return pd.DataFrame({"col": [1, 2, 3]})

    # Register first function
    @cache.register_as("test_table")
    def first_func() -> pd.DataFrame:
        return pd.DataFrame({"col": [1, 2, 3]})

    # Try to register another function with same table name
    with pytest.raises(
        ValueError, match="Table name 'test_table' is already registered"
    ):

        @cache.register_as("test_table")
        def second_func() -> pd.DataFrame:
            return pd.DataFrame({"col": [4, 5, 6]})


def test_duck_loader_error_handling(temp_db):
    """Test error handling in DuckLoader."""
    # Create a DuckLoader instance
    loader = duckloader_factory(temp_db)

    # Test accessing non-existent table
    with pytest.raises(AttributeError):
        _ = loader.non_existent_table

    # Test accessing table with invalid name
    with pytest.raises(AttributeError):
        _ = loader.invalid_table_name


def test_run_query_df_polars(temp_db):
    """Test run_query_df returns polars DataFrame when use_polars=True."""
    cache = DuckCacher(temp_db, use_polars=True)

    @cache.register
    def test_table() -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3]})

    # Trigger table creation
    _ = cache.df_test_table
    # Now run a query
    result = cache.run_query_df("SELECT * FROM main.test_table")
    assert isinstance(result, pl.DataFrame)
    assert result.shape == (3, 1)
    assert result.columns == ["a"]


def test_needs_sync_polars(temp_db):
    """Test _needs_sync with a polars DataFrame result."""
    cache = DuckCacher(temp_db)
    # Insert a row into duckcacher_state using polars
    df = pl.DataFrame({"code_hash": ["abc"]})
    # Patch run_query_df to return a polars DataFrame
    cache.run_query_df = lambda *a, **k: df
    # Should return True if hash doesn't match
    assert cache._needs_sync("main", "func", "not_abc")
    # Should return False if hash matches
    assert not cache._needs_sync("main", "func", "abc")


def test_duckloader_factory_polars(tmp_path):
    """Test duckloader_factory returns polars DataFrame when use_polars=True."""
    db_path = tmp_path / "test_polars.db"
    cache = DuckCacher(str(db_path))

    @cache.register
    def test_table() -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3]})

    _ = cache.df_test_table
    loader = duckloader_factory(str(db_path), use_polars=True)
    df = loader.df_test_table
    assert isinstance(df, pl.DataFrame)
    assert df.shape == (3, 1)
    assert df.columns == ["a"]


def test_sync_nonexistent_table(temp_db):
    """Test sync raises ValueError when trying to sync a non-existent table."""
    cache = DuckCacher(temp_db)

    # Register a table first so we can see it in the error message
    @cache.register_as("test_table")
    def test_func() -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3]})

    with pytest.raises(
        ValueError,
        match="No registered function found for table 'nonexistent_table'. Valid tables are: test_table",
    ):
        cache.sync("nonexistent_table")


def test_register_as_duplicate(temp_db):
    """Test register_as raises ValueError for a duplicate table name."""
    cache = DuckCacher(temp_db)

    @cache.register_as("test_table")
    def first_func() -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3]})

    with pytest.raises(
        ValueError, match="Table name 'test_table' is already registered"
    ):

        @cache.register_as("test_table")
        def second_func() -> pd.DataFrame:
            return pd.DataFrame({"a": [4, 5, 6]})


def test_sync_nonexistent_attribute(temp_db):
    """Test sync raises ValueError when accessing a non-existing attribute."""
    cache = DuckCacher(temp_db)
    with pytest.raises(ValueError, match="No registered function found for table"):
        cache.sync("df_not_existing")


def test_sync_single_table(temp_db):
    """Test that sync correctly updates a single table and maintains data integrity."""
    cache = DuckCacher(temp_db)

    # Register a function that returns a simple dataframe
    @cache.register
    def foo() -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3]})

    # First access should create the table
    df1 = cache.df_foo
    pd.testing.assert_frame_equal(df1, pd.DataFrame({"a": [1, 2, 3]}))

    # Modify the function to return different data
    @cache.register
    def foo() -> pd.DataFrame:
        return pd.DataFrame({"a": [4, 5, 6]})

    # Sync should update the table with new data
    cache.sync("foo")

    # Verify the data was updated
    df2 = cache.df_foo
    pd.testing.assert_frame_equal(df2, pd.DataFrame({"a": [4, 5, 6]}))

    # Verify we get copies of the dataframe
    assert df1 is not df2, "Should get different dataframe objects"

    # Verify the dataframes are not equal after modification
    with pytest.raises(AssertionError):
        pd.testing.assert_frame_equal(df1, df2)

    # Test sync without table name (should sync all tables)
    @cache.register
    def bar() -> pd.DataFrame:
        return pd.DataFrame({"b": [7, 8, 9]})

    # First access should create the table
    df3 = cache.df_bar
    pd.testing.assert_frame_equal(df3, pd.DataFrame({"b": [7, 8, 9]}))

    # Modify both functions
    @cache.register
    def foo() -> pd.DataFrame:
        return pd.DataFrame({"a": [10, 11, 12]})

    @cache.register
    def bar() -> pd.DataFrame:
        return pd.DataFrame({"b": [13, 14, 15]})

    # Sync all tables
    cache.sync()

    # Verify both tables were updated
    df4 = cache.df_foo
    df5 = cache.df_bar

    pd.testing.assert_frame_equal(df4, pd.DataFrame({"a": [10, 11, 12]}))
    pd.testing.assert_frame_equal(df5, pd.DataFrame({"b": [13, 14, 15]}))

    # Verify we get copies of the dataframes
    assert df3 is not df5, "Should get different dataframe objects"

    # Verify the dataframes are not equal after modification
    with pytest.raises(AssertionError):
        pd.testing.assert_frame_equal(df3, df5)


def test_register_with_args(temp_db):
    """Test that register raises ValueError for functions with arguments."""
    cache = DuckCacher(temp_db)

    # Test function with positional args
    with pytest.raises(
        ValueError, match="Function 'func_with_args' cannot have any arguments"
    ):

        @cache.register
        def func_with_args(x, y):
            return pd.DataFrame({"col": [x, y]})

    # Test function with kwargs
    with pytest.raises(
        ValueError, match="Function 'func_with_kwargs' cannot have any arguments"
    ):

        @cache.register
        def func_with_kwargs(x=1, y=2):
            return pd.DataFrame({"col": [x, y]})

    # Test function with *args
    with pytest.raises(
        ValueError, match="Function 'func_with_star_args' cannot have any arguments"
    ):

        @cache.register
        def func_with_star_args(*args):
            return pd.DataFrame({"col": args})

    # Test function with **kwargs
    with pytest.raises(
        ValueError, match="Function 'func_with_star_kwargs' cannot have any arguments"
    ):

        @cache.register
        def func_with_star_kwargs(**kwargs):
            return pd.DataFrame({"col": list(kwargs.values())})


def test_register_as_with_args(temp_db):
    """Test that register_as raises ValueError for functions with arguments."""
    cache = DuckCacher(temp_db)

    # Test function with positional args
    with pytest.raises(
        ValueError, match="Function 'func_with_args' cannot have any arguments"
    ):

        @cache.register_as("table1")
        def func_with_args(x, y):
            return pd.DataFrame({"col": [x, y]})

    # Test function with kwargs
    with pytest.raises(
        ValueError, match="Function 'func_with_kwargs' cannot have any arguments"
    ):

        @cache.register_as("table2")
        def func_with_kwargs(x=1, y=2):
            return pd.DataFrame({"col": [x, y]})

    # Test function with *args
    with pytest.raises(
        ValueError, match="Function 'func_with_star_args' cannot have any arguments"
    ):

        @cache.register_as("table3")
        def func_with_star_args(*args):
            return pd.DataFrame({"col": args})

    # Test function with **kwargs
    with pytest.raises(
        ValueError, match="Function 'func_with_star_kwargs' cannot have any arguments"
    ):

        @cache.register_as("table4")
        def func_with_star_kwargs(**kwargs):
            return pd.DataFrame({"col": list(kwargs.values())})


def test_duckloader_factory_ls_isolation(tmp_path):
    """Test that ls() returns different tables for different database files."""
    # Create two different database files
    db1_path = tmp_path / "test1.db"
    db2_path = tmp_path / "test2.db"

    # Create two different cachers with different tables
    cache1 = DuckCacher(str(db1_path))
    cache2 = DuckCacher(str(db2_path))

    # Register different tables in each cache
    @cache1.register
    def table1():
        return pd.DataFrame({"col1": [1, 2, 3]})

    @cache1.register
    def table2():
        return pd.DataFrame({"col2": [4, 5, 6]})

    @cache2.register
    def table1():
        return pd.DataFrame({"col3": [7, 8, 9]})

    @cache2.register
    def table3():
        return pd.DataFrame({"col4": [10, 11, 12]})

    # Sync both caches to ensure tables are created
    cache1.sync()
    cache2.sync()

    # Create loaders for each database
    loader1 = duckloader_factory(str(db1_path))
    loader2 = duckloader_factory(str(db2_path))

    # Get the list of tables from each loader
    tables1 = set(loader1.ls())
    tables2 = set(loader2.ls())

    # Verify the tables are different
    assert tables1 == {
        "df_table1",
        "df_table2",
    }, "First loader should only see its own tables"
    assert tables2 == {
        "df_table1",
        "df_table3",
    }, "Second loader should only see its own tables"

    # Verify there's no overlap in the actual data
    df1_table1 = loader1.df_table1
    df2_table1 = loader2.df_table1

    assert df1_table1["col1"].tolist() == [
        1,
        2,
        3,
    ], "First loader's table1 should have col1"
    assert df2_table1["col3"].tolist() == [
        7,
        8,
        9,
    ], "Second loader's table1 should have col3"


def test_direct_function_call_populates_cache_and_db(tmp_path):
    """Test that calling a registered function directly populates cache and database."""
    db_path = tmp_path / "test.db"
    cache = DuckCacher(str(db_path))

    @cache.register
    def my_table():
        return pd.DataFrame({"a": [1, 2, 3]})

    # Call the function directly
    df = my_table()

    # The in-memory cache should have the table
    assert "main.my_table" in cache._cache

    # The DuckDB table should exist
    with duckdb.connect(str(db_path)) as conn:
        tables = conn.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = 'main'"
        ).fetchall()
        table_names = {row[0] for row in tables}
        assert "my_table" in table_names

    # Verify the data is correct
    expected_df = pd.DataFrame({"a": [1, 2, 3]})
    pd.testing.assert_frame_equal(df, expected_df)


def test_duckcacher_attribute_isolation(tmp_path):
    """Test that different DuckCacher instances don't interfere with each other's attributes."""
    # Create two different database files
    db1_path = tmp_path / "test1.db"
    db2_path = tmp_path / "test2.db"

    # Create two different cachers
    cache1 = DuckCacher(str(db1_path))
    cache2 = DuckCacher(str(db2_path))

    # Register different functions with the same name in each cache
    @cache1.register
    def table1():
        return pd.DataFrame({"col1": [1, 2, 3]})

    @cache1.register
    def table2():
        return pd.DataFrame({"col2": [4, 5, 6]})

    @cache2.register
    def table1():
        return pd.DataFrame({"col3": [7, 8, 9]})

    @cache2.register
    def table3():
        return pd.DataFrame({"col4": [10, 11, 12]})

    # Access the tables via df_ attributes
    df1_table1 = cache1.df_table1
    df1_table2 = cache1.df_table2
    df2_table1 = cache2.df_table1
    df2_table3 = cache2.df_table3

    # Verify the data is different and correct
    assert df1_table1["col1"].tolist() == [
        1,
        2,
        3,
    ], "First cache's table1 should have col1"
    assert df1_table2["col2"].tolist() == [
        4,
        5,
        6,
    ], "First cache's table2 should have col2"
    assert df2_table1["col3"].tolist() == [
        7,
        8,
        9,
    ], "Second cache's table1 should have col3"
    assert df2_table3["col4"].tolist() == [
        10,
        11,
        12,
    ], "Second cache's table3 should have col4"

    # Verify that each cache only has its own registrations
    assert set(cache1._registrations.keys()) == {"table1", "table2"}
    assert set(cache2._registrations.keys()) == {"table1", "table3"}

    # Verify that accessing non-existent attributes raises AttributeError
    with pytest.raises(AttributeError):
        _ = cache1.df_table3

    with pytest.raises(AttributeError):
        _ = cache2.df_table2


def test_register_as_wrapped_function_call(tmp_path):
    """Test that calling the wrapped function from register_as directly works and covers line 347."""
    db_path = tmp_path / "test.db"
    cache = DuckCacher(str(db_path))

    @cache.register_as("my_table")
    def some_function():
        return pd.DataFrame({"a": [1, 2, 3]})

    # Call the wrapped function directly (this should trigger line 347)
    df = some_function()

    # Verify the data is correct
    expected_df = pd.DataFrame({"a": [1, 2, 3]})
    pd.testing.assert_frame_equal(df, expected_df)

    # Verify the table was created in the database
    with duckdb.connect(str(db_path)) as conn:
        tables = conn.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = 'main'"
        ).fetchall()
        table_names = {row[0] for row in tables}
        assert (
            "some_function" in table_names
        ), "Table should be created with the original function name"

    # Verify the dataframe is in the cache
    assert (
        "main.some_function" in cache._cache
    ), "Dataframe should be in the cache after calling wrapped function"

    # Verify we can also access it via the df_ attribute
    df2 = cache.df_my_table
    pd.testing.assert_frame_equal(df2, expected_df)



from easier.dataframe_tools import (
    pandas_time_to_utc_seconds,
    pandas_utc_seconds_to_time,
    slugify,
)
from unittest import TestCase
import datetime
import pandas as pd


class TestSlugify(TestCase):

    def test_strings(self):
        text = "hello:  world "
        camel_text = "helloWorld"
        out_text = slugify(text)
        out_text2 = slugify(text, sep=".")
        self.assertEqual(out_text, "hello_world")
        self.assertEqual(out_text2, "hello.world")
        self.assertEqual(slugify(camel_text, kill_camel=True), "hello_world")

    def test_iterable(self):
        values = [
            " these ",
            "are_things",
            "ThatNeedToBe",
            "*@#$%^fixed;#$ .   rightNow",
        ]
        fixed = slugify(values, kill_camel=True)
        self.assertListEqual(
            fixed, ["these", "are_things", "that_need_to_be", "fixed_right_now"]
        )


class TestUnixTimeConversion(TestCase):

    def setUp(self):
        dates = pd.date_range("1/1/2020", "1/10/2020")
        df = pd.DataFrame({"time1": dates.values})
        df["time2"] = df.time1
        self.df = df

    def test_conversion(self):
        dfin = self.df.copy()
        self.assertTrue(all(["datetime64" in str(val) for val in dfin.dtypes.values]))
        df = pandas_time_to_utc_seconds(dfin)
        self.assertTrue(all(["int64" in str(val) for val in df.dtypes.values]))
        df_back = pandas_utc_seconds_to_time(df)
        self.assertEqual((dfin - df_back).max().max().days, 0)
        df = pandas_time_to_utc_seconds(dfin, columns=["time1"])
        self.assertTrue(any(["int64" in str(val) for val in df.dtypes.values]))
        self.assertTrue(any(["datetime64" in str(val) for val in df.dtypes.values]))
        df["time1"] = pandas_utc_seconds_to_time(df.time1)
        self.assertEqual((dfin - df).max().max().days, 0)
        df = pandas_time_to_utc_seconds(dfin, ["time2"])
        epoch = datetime.datetime(1970, 1, 1)
        expected_seconds = (df.time1.iloc[0] - epoch).total_seconds()
        computed_seconds = df.time2.iloc[0]
        self.assertEqual(expected_seconds, computed_seconds)
        with self.assertRaises(ValueError):
            pandas_utc_seconds_to_time(df.time1, columns="time")
        with self.assertRaises(ValueError):
            pandas_utc_seconds_to_time([1, 2, 3])


def test_hex_dataframe_round_trip():
    import pandas as pd
    from easier.dataframe_tools import hex_from_dataframe, hex_to_dataframe

    df = pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"], "C": [1.1, 2.2, 3.3]})
    hex_str = hex_from_dataframe(df)
    df2 = hex_to_dataframe(hex_str)
    pd.testing.assert_frame_equal(df, df2)


def test_duckdb_hex_round_trip():
    import duckdb
    import pandas as pd
    from easier.dataframe_tools import hex_from_duckdb, hex_to_duckdb
    # Create an in-memory DuckDB and a test table
    conn = duckdb.connect(":memory:")
    df = pd.DataFrame({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    conn.execute("CREATE TABLE test AS SELECT * FROM df")
    # Export to hex
    hex_dump = hex_from_duckdb(conn)
    # Import back from hex
    new_conn = hex_to_duckdb(hex_dump)
    # Fetch the data from the new connection
    result_df = new_conn.execute("SELECT * FROM test").fetchdf()
    # Check that the data matches
    pd.testing.assert_frame_equal(result_df, df)



# from google.cloud import bigquery
# from google.oauth2 import service_account
# from typing import Optional
# import abc
# import easier as ezr
# import numpy as np
# import os
# import pandas as pd
# import pandas_gbq

# import abc
# import os
# import easier as ezr
# from typing import Optional
# import pandas_gbq
# import pandas as pd
# import numpy as np
# example = '\nfrom easier.bigquery import BQAppender, BQTable, BQDataset\nimport pandas as pd\nimport datetime\n\n\nclass MyAppender(BQAppender):\n    def get_dataframe(self, **kwargs):\n        # self.validate_kwargs(**kwargs)\n\n        df = pd.DataFrame(\n            {\n                \'my_str\': [\'a\', \'b\'],\n                \'my_int\': [1, 2],\n                \'my_float\': [3., 4.],\n            }\n        )\n        time = kwargs[\'time\']\n        df[\'time\'] = time\n        return df\n\n    def validate_kwargs(self, **kwargs):\n        if \'time\' not in kwargs:\n            raise ValueError(\'kwargs must contain "time"\')\n\n\nclass MyDataset(BQDataset):\n    dataset_id = \'rob\'\n\n    my_table = BQTable(\n        name=\'my_table\',\n        appender_class=MyAppender,\n        schema={\n            \'time\': \'TIMESTAMP\',\n            \'my_str\': \'STRING\',\n            \'my_int\': \'INTEGER\',\n            \'my_float\': \'FLOAT\',\n        }\n\n    )\n\n    def push(self):\n        time = datetime.datetime.now()\n        for table in self.tables:\n            table.append(time=time)\n\n\nproject_id = \'ambition-analytics\'\nds = MyDataset(project_id, create_missing_dataset=True)\nds.push()\n'

# class BQAppender(metaclass=abc.ABCMeta):
#     """
#     A BQTable instance needs to know how to grab data to append to the database.
#     One of the arguments it takes is a appender class.  This is the base class you should
#     derive from when creating your own Appenders.
#     """

#     @abc.abstractmethod
#     def get_dataframe(self, **kwargs):
#         """
#         This method must be implemented.  It should contain the implementation of how
#         you want to grab the data you want to push into the database.

#         Any kwargs can be supplied to this method.  If you wish to validate the kwargs to this method,
#         you can do so using the validate_kwargs(*kwargs) method.
#         """

#     def validate_kwargs(self, **kwargs):
#         """
#         Override this method to provide (optional) validation on the kwargs sent to the finalize_frame()
#         method.
#         """

# class BQTable:
#     """
#     BQTable is a descriptor you should use when defining your BQDataset class.  It provides an
#     interface for appending and accessing data from a particular table in your Bigquery Dataset.
#     """
#     _PARTITIONING_TYPE = 'DAY'
#     _TYPE_LOOKUP = {pd.Timestamp: 'TIMESTAMP', np.int64: 'INTEGER', np.float64: 'FLOAT', str: 'STRING'}

#     def __init__(self, name: str, schema: dict, appender_class: Optional[BQAppender]=None):
#         """
#         Args:
#             name: the name of the databse table to create/use in Bigquery
#             schema: A dictionary with keys of field names and values of field types.
#             appender_class: The class to use for appending data.
#         """
#         self.name = name
#         self.schema = schema
#         self.existance_confirmed = False
#         self.appender_class = appender_class

#     def __str__(self):
#         return f'BQTable({self.name})'

#     def __repr__(self):
#         return self.__str__()

#     def _table_exists(self, dataset):
#         """
#         Returns whether or not this table exists
#         """
#         return self.name in [t.table_id for t in dataset.client.list_tables(dataset.dataset_id)]

#     def _ensure_exists(self, dataset):
#         """
#         If it doesn't already exist, create this table on the provided dataset using the defined schema.
#         """
#         if not self._table_exists(dataset):
#             table_ref = dataset.dataset.table(self.name)
#             schema = list((bigquery.SchemaField(field, dtype) for field, dtype in self.schema.items()))
#             table = bigquery.Table(table_ref, schema=schema)
#             table.partitioning_type = self._PARTITIONING_TYPE
#             dataset.client.create_table(table)
#         self.existance_confirmed = True

#     def __get__(self, dataset, dataset_class):
#         """
#         This is what makes this class a descriptor
#         """
#         self._dataset = dataset
#         self._ensure_exists(dataset)
#         return self

#     def _check_dataframe(self, df):
#         """
#         Checks to make sure a dataframe is valid for appending to the bigquery table
#         """
#         if df.empty:
#             return
#         expected_cols = set(self.schema.keys())
#         received_cols = set(df.columns)
#         missing_cols = expected_cols - received_cols
#         if missing_cols:
#             raise ValueError(f'Dataframe missing the following columns: {missing_cols}')
#         bad_list = []
#         for col, col_type in self.schema.items():
#             mapped_type = self._TYPE_LOOKUP.get(type(df[col].iloc[0]))
#             if col_type != mapped_type:
#                 bad_list.append(col)
#         if bad_list:
#             raise ValueError(f'Bad columns types for these columns: {bad_list}')

#     def append_dataframe(self, df):
#         """
#         Call this method if you have a dataframe you want to manually append to the table.  Usually
#         you won't want to use this method, but instead use the .append() method which will use your
#         defined Appender class to create the dataframe and then append it to the bigquery table.
#         """
#         if df.empty:
#             return
#         self._check_dataframe(df)
#         df = df[list(self.schema.keys())]
#         pandas_gbq.to_gbq(df, self.name_in_project, project_id=self._dataset.project_id, if_exists='append', progress_bar=False)

#     def get_dataframe(self, **kwargs):
#         """
#         This will pull in the source dataframe from this tables appender.  This is the dataframe
#         that is used as the source for appending data to the bigquery table.
#         """
#         if self.appender_class is None:
#             raise ValueError('You must provide an appender class to the constructor if you want to call append()')
#         appender = self.appender_class()
#         appender.validate_kwargs(**kwargs)
#         df = appender.get_dataframe(**kwargs)
#         return df

#     def append(self, dry_run=False, **kwargs):
#         """
#         This is the preferred method for appending data to the bigquery database.  It will
#         use the supplied Appender class to load a frame whose data will be pushed to bigquery.
#         You can supply optional keyword arguments that will be passed to the .finalize_frame() method
#         of your appender class.  These kwargs provide the ability to add whatever final/custom processing
#         you want to your dataframe.
#         """
#         df = self.get_dataframe(**kwargs)
#         if not dry_run:
#             self.append_dataframe(df)

#     @ezr.cached_property
#     def full_name(self):
#         """
#         This is the full name of the table.  It is what you should use in the
#         FROM clause of any sql query.
#         """
#         return f'{self._dataset.project_id}.{self.name_in_project}'

#     @ezr.cached_property
#     def name_in_project(self):
#         """
#         The name of this table within the project
#         """
#         return f'{self._dataset.dataset_id}.{self.name}'

#     @ezr.cached_property
#     def latest_partition_time(self):
#         """
#         An attribute with a cached latest parition time
#         """
#         return self.get_latest_partition_time()

#     @ezr.cached_container
#     def latest_partition_frame(self):
#         return self.get_latest_partition_frame()

#     def get_latest_partition_frame(self):
#         """
#         An attribute with a cached dataframe of the latest parition data
#         """
#         sql = '\n            SELECT\n                *\n            FROM\n                {table}\n            WHERE\n                _PARTITIONDATE = "{partition_date_iso}"\n        '.format(table=self.full_name, partition_date_iso=self.latest_partition_time.date().isoformat())
#         df = self._dataset.query(sql)
#         return df

#     def get_latest_partition_time(self):
#         """
#         Grabs the most recent partition time for the table
#         """
#         sql_template = '\n            SELECT\n                _PARTITIONDATE AS partition_date\n            FROM\n                {full_name}\n            WHERE\n                _PARTITIONDATE >= DATE_SUB(CURRENT_DATE(), INTERVAL {days} DAY)\n            ORDER BY\n                partition_date DESC\n            LIMIT 1\n        '
#         days = 1
#         while days <= 256:
#             sql = sql_template.format(full_name=self.full_name, days=days)
#             df = self._dataset.query(sql)
#             if not df.empty:
#                 return df.partition_date.iloc[0].to_pydatetime()
#             days = 2 * days
#         raise RuntimeError("To reduce data-size you can't query partitioned more than 256 days ago")

#     def get_latest_data(self, time_field_name):
#         """
#         If you have timestamped data in the database, this will grab all data with the latest timestamp
#         for the latest partition.
#         """
#         sql = '\n            WITH latest_partition AS (\n                SELECT\n                    *\n                FROM\n                    {table}\n                WHERE\n                    _PARTITIONDATE = "{partition_date_iso}"\n            )\n            SELECT\n                *\n            FROM\n                latest_partition\n            WHERE\n                {time_field_name} = (SELECT MAX({time_field_name}) FROM latest_partition)\n        '.format(table=self.full_name, time_field_name=time_field_name, partition_date_iso=self.latest_partition_time.date().isoformat())
#         df = self._dataset.query(sql)
#         return df

# class BQDataset:

#     def __init__(self, project_id, *, dataset_id=None, create_missing_dataset=False, **kwargs):
#         """
#         A class that knows how to define and work with tables in a BigQuery dataset.

#         Args:
#             project_id: The project id this dataset belongs to
#             dataset_id: The (optional) id for the dataset
#             create_missing_dataset: Setting this flag will create non-existing datasets.
#             kwargs: Any arguments you pass here will get passed to a .constructor_kwargs attribute.
#         """
#         if dataset_id is not None:
#             self.dataset_id = dataset_id
#         self.project_id = project_id
#         self.constructor_kwargs = kwargs
#         if not hasattr(self, 'dataset_id'):
#             raise ValueError('You must set dataset_id as class-variable on your derived class or pass data_set_id to the constructor')
#         self._ensure_dataset_exists(create_missing_dataset)

#     @ezr.cached_property
#     def full_dataset_name(self):
#         return f'{self.project_id}.{self.dataset_id}'

#     @ezr.cached_property
#     def all_dataset_names(self):
#         return [d.dataset_id for d in self.client.list_datasets()]

#     def _ensure_dataset_exists(self, create_missing_dataset):
#         if self.dataset_id not in self.all_dataset_names:
#             if create_missing_dataset:
#                 self._create_dataset()
#             else:
#                 msg = f'Dataset "{self.dataset_id}" does not exist. You can set create_missing_dataset=True to create it'
#                 raise RuntimeError(msg)

#     def ensure_all_tables_exist(self):
#         for table in self.tables:
#             table._ensure_exists(self)

#     def _create_dataset(self):
#         dataset = bigquery.Dataset(self.full_dataset_name)
#         dataset.location = 'US'
#         self.client.create_dataset(dataset, timeout=30)

#     @property
#     def auth_file(self):
#         auth_file = None
#         config_dir = os.path.expanduser('~/.config/gbq')
#         if os.path.isdir(config_dir):
#             auth_file_list = os.listdir(config_dir)
#             if config_dir:
#                 auth_file = os.path.join(config_dir, auth_file_list[0])
#         return auth_file

#     @classmethod
#     def get_table_names(cls):
#         """
#         A method to return all table names
#         """
#         table_names = []
#         for name, obj in vars(cls).items():
#             if isinstance(obj, BQTable):
#                 table_names.append(name)
#         return table_names

#     @ezr.cached_property
#     def tables(self):
#         """
#         A list attribute containing all tables
#         """
#         tables = []
#         for name in self.get_table_names():
#             tables.append(getattr(self, name))
#         return tables

#     @property
#     def on_dev_machine(self):
#         """
#         This will probably change when I do this for real, but for now, I check if I'm
#         in a cloud function or not by checking for the existence of an auth file
#         """
#         return self.auth_file is not None

#     @ezr.cached_property
#     def client(self):
#         """
#         An attribute holding the google client
#         """
#         client_kwargs = {'project': self.project_id}
#         if self.on_dev_machine:
#             client_kwargs.update(credentials=service_account.Credentials.from_service_account_file(self.auth_file))
#         client = bigquery.Client(**client_kwargs)
#         return client

#     @ezr.cached_property
#     def dataset(self):
#         """
#         An attribute holding the google dataset
#         """
#         return self.client.get_dataset(self.full_dataset_name)

#     def query(self, sql_query, show_progress=False):
#         """
#         Run BigTable queries on this dataset. Tables can be referenced in FROM and JOIN SQL clauses
#         using their .full_name attributes injected into a query template.
#         """
#         if show_progress:
#             progress_bar_type = 'tqdm'
#         else:
#             progress_bar_type = None
#         df = pandas_gbq.read_gbq(sql_query, project_id=self.project_id, progress_bar_type=progress_bar_type)
#         return df



import pandas as pd
import duckdb
import hashlib
import inspect
from typing import Any, Callable, Dict
import os


class _ManagedConnection:
    """A class that manages DuckDB connections with different behaviors for file and memory databases.

    For file-based databases, it creates new connections on each context entry.
    For in-memory databases, it maintains a single persistent connection.
    """

    def __init__(self, file_name: str):
        self.file_name = file_name
        self._memory_conn = (
            None if file_name != ":memory:" else duckdb.connect(file_name)
        )

    def __enter__(self):
        if self.file_name == ":memory:":
            return self._memory_conn
        return duckdb.connect(self.file_name)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.file_name != ":memory:":
            # Only close file-based connections
            return True  # Suppress any exceptions

    @property
    def connection(self):
        """Get the current connection. For memory databases, returns the persistent connection.
        For file databases, creates a new connection."""
        if self.file_name == ":memory:":
            return self._memory_conn
        return duckdb.connect(self.file_name)


class Duckmirror:
    """A class that provides automatic synchronization between Python functions and DuckDB tables.

    Duckmirror automatically tracks changes to decorated functions and their corresponding
    DuckDB tables. It maintains a cache of DataFrames in memory (optional) and provides
    automatic synchronization between function outputs and cached database tables.

    The class supports both standalone functions and class methods as data sources.
    When used with class methods, it creates a schema named after the class and stores
    tables within that schema.

    Attributes:
        file_name (str): Path to the DuckDB database file or ":memory:" for in-memory database
        cache_in_memory (bool): Whether to maintain an in-memory cache of DataFrames
        _cache (Dict[str, pd.DataFrame]): Internal cache of DataFrames
        _registrations (Dict[str, Callable]): Internal registry of decorated functions
        _conn_manager (_ManagedConnection): Connection manager instance
    """

    def __init__(
        self,
        file_name: str = ":memory:",
        overwrite: bool = False,
        cache_in_memory: bool = True,
    ):
        """Initialize a new Duckmirror instance.

        Args:
            file_name (str, optional): Path to the DuckDB database file. Use ":memory:" for
                an in-memory database. Defaults to ":memory:".
            overwrite (bool, optional): Whether to overwrite existing database file.
                If True and file exists, it will be deleted. Defaults to False.
            cache_in_memory (bool, optional): Whether to maintain an in-memory cache of
                DataFrames for faster access. Defaults to True.
        """
        self.file_name = file_name
        self.cache_in_memory = cache_in_memory
        self._cache: Dict[str, pd.DataFrame] = {}
        self._registrations: Dict[str, Callable] = {}

        # Overwrite logic: delete file if overwrite is True and not in-memory
        if overwrite and file_name != ":memory:":
            if os.path.exists(file_name):
                os.remove(file_name)

        # Initialize connection manager
        self._conn_manager = _ManagedConnection(file_name)

        # Create duckmirror schema and state table if they don't exist
        with self._conn_manager as conn:
            conn.execute("CREATE SCHEMA IF NOT EXISTS duckmirror")
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS duckmirror.duckmirror_state (
                    schema_name VARCHAR,
                    function_name VARCHAR,
                    code_hash VARCHAR,
                    PRIMARY KEY (schema_name, function_name)
                )
            """
            )

    @property
    def conn(self):
        """Get the current database connection."""
        return self._conn_manager.connection

    def run_query_df(self, query: str, *args, **kwargs) -> pd.DataFrame:
        """Execute a SELECT query and return the result as a DataFrame."""
        with self._conn_manager as conn:
            result = conn.execute(query, parameters=args if args else None)
            return result.df()

    def run_execute(self, query: str, *args, **kwargs):
        """Execute a query that does not return a DataFrame (DDL/DML)."""
        with self._conn_manager as conn:
            conn.execute(query, parameters=args if args else None)

    def _get_code_hash(self, func: Callable) -> str:
        """Get the hash of a function's source code."""
        source = inspect.getsource(func)
        return hashlib.sha256(source.encode()).hexdigest()

    def _needs_sync(self, schema: str, func_name: str, current_hash: str) -> bool:
        """Check if a function needs to be synced based on its hash."""
        df = self.run_query_df(
            """
            SELECT code_hash 
            FROM duckmirror.duckmirror_state 
            WHERE schema_name = ? AND function_name = ?
        """,
            schema,
            func_name,
        )
        if df.empty:
            return True
        return df.iloc[0]["code_hash"] != current_hash

    def _ensure_schema(self, schema: str) -> None:
        """Ensure a schema exists in the database."""
        self.run_execute(f"CREATE SCHEMA IF NOT EXISTS {schema}")

    def _store_dataframe(self, schema: str, table: str, df: pd.DataFrame) -> None:
        """Store a dataframe in the database and optionally in memory."""
        self._ensure_schema(schema)
        with self._conn_manager as conn:
            conn.execute(f"DROP TABLE IF EXISTS {schema}.{table}")
            conn.register("df", df)
            conn.execute(f"CREATE TABLE {schema}.{table} AS SELECT * FROM df")
        cache_key = f"{schema}.{table}"
        if self.cache_in_memory:
            self._cache[cache_key] = df

    def _get_cached_dataframe(self, schema: str, table: str) -> pd.DataFrame:
        """Get a dataframe from cache or database."""
        cache_key = f"{schema}.{table}"
        if self.cache_in_memory and cache_key in self._cache:
            return self._cache[cache_key].copy()
        return self.run_query_df(f"SELECT * FROM {schema}.{table}")

    def register(self, func: Callable) -> Callable:
        """Register a function or method to be synchronized with a DuckDB table."""
        if "." in func.__qualname__:
            # It's a method: register for both class property and Duckmirror lookup
            key = func.__qualname__.replace(".", "_")
            self._registrations[key] = func
            # Also register with just the function name for class property access
            self._registrations[func.__name__] = func
            return property(func)
        else:
            # Standalone function
            key = func.__name__
            self._registrations[key] = func
            return func

    def __getattr__(self, name: str) -> Any:
        """Handle dynamic attribute access for df_<name>."""
        if name.startswith("df_"):
            func_name = name[3:]  # Remove 'df_' prefix
            # Try to find the function by its key
            if func_name in self._registrations:
                func = self._registrations[func_name]
                # Use qualname to distinguish method vs standalone
                if "." in func.__qualname__ and func.__qualname__.count(".") == 1:
                    schema = func.__qualname__.split(".")[0]
                    table_name = func.__name__
                elif "." in func.__qualname__:
                    # If function is nested, treat as standalone
                    schema = "main"
                    table_name = func.__name__
                else:
                    schema = "main"
                    table_name = func.__name__
                instance = None
            else:
                raise AttributeError(
                    f"'{self.__class__.__name__}' has no attribute '{name}'"
                )

            # Get current hash
            current_hash = self._get_code_hash(func)

            # Check if sync is needed
            if self._needs_sync(schema, table_name, current_hash):
                # Run the function
                if instance is not None:
                    df = func(instance)
                else:
                    df = func()

                # Store the result
                self._store_dataframe(schema, table_name, df)

                # Update hash
                self.run_execute(
                    """
                    INSERT OR REPLACE INTO duckmirror.duckmirror_state 
                    (schema_name, function_name, code_hash)
                    VALUES (?, ?, ?)
                """,
                    schema,
                    table_name,
                    current_hash,
                )

            return self._get_cached_dataframe(schema, table_name)
        raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{name}'")



from collections import namedtuple

from pathlib import Path
from typing import List, Dict, Any, Tuple
import copy
import easier as ezr
import functools
import importlib

import os
import urllib.parse


def pg_creds_from_env(kind="dict", force_docker=False):
    """
    Pulls PostgreSQL credentials from environment variables or defaults to Docker settings.

    This function retrieves database connection credentials from environment variables.
    If environment variables are not set, it defaults to standard Docker PostgreSQL
    credentials. The credentials can be returned either as a dictionary or as a
    connection URL.

    Args:
        kind (str, optional): The format to return credentials in. Must be either
            "dict" or "url". Defaults to "dict".
        force_docker (bool, optional): If True, ignores environment variables and
            returns Docker default credentials. Defaults to False.

    Returns:
        Union[dict, str]: If kind="dict", returns a dictionary with connection
            parameters. If kind="url", returns a PostgreSQL connection URL string.

    Raises:
        ValueError: If kind is not one of the allowed values ("dict" or "url").
    """
    allowed_kinds = ["url", "dict"]
    if kind not in allowed_kinds:
        raise ValueError(f"Allowed kinds are {allowed_kinds}")
    if force_docker:
        env = {}
    else:
        env = os.environ.copy()
    creds = {
        "host": env.get("PGHOST", "db"),
        "port": env.get("PGPORT", "5432"),
        "database": env.get("PGDATABASE", "postgres"),
        "user": env.get("PGUSER", "postgres"),
        "password": env.get("PGPASSWORD", "postgres"),
    }
    encoded_user = urllib.parse.quote(creds["user"])
    encoded_password = urllib.parse.quote(creds["password"])
    url = f"postgresql://{encoded_user}:{encoded_password}@{creds['host']}:{creds['port']}/{creds['database']}"
    if kind == "dict":
        return creds
    else:
        return url


class PG:
    """
    A PostgreSQL database connection and query execution class.

    This class provides a convenient interface for connecting to PostgreSQL databases,
    executing queries, and retrieving results in various formats (DataFrame, tuples,
    dictionaries, etc.). It supports both direct connection parameters and Django
    database configurations.

    Args:
        host (str, optional): Database host. Defaults to PGHOST environment variable.
        user (str, optional): Database user. Defaults to PGUSER environment variable.
        password (str, optional): Database password. Defaults to PGPASSWORD environment variable.
        dbname (str, optional): Database name. Defaults to PGDATABASE environment variable.
        use_django (bool, optional): If True, use Django database settings. Defaults to False.

    Attributes:
        _sql (str): The SQL query to be executed
        _context (dict): Context for query execution
        _conn_kwargs (dict): Database connection parameters
        _raw_results (list): Raw query results
        _raw_columns (list): Column names from query results
    """

    _sql = None
    _context = None
    _conn_kwargs = None
    _raw_results = None
    _raw_columns = None

    def __init__(self, **kwargs):
        """
        Initialize a PostgreSQL database connection.

        Args:
            **kwargs: Connection parameters that can include:
                - host: Database host (default from PGHOST env var)
                - user: Database user (default from PGUSER env var)
                - password: Database password (default from PGPASSWORD env var)
                - dbname: Database name (default from PGDATABASE env var)
                - use_django: If True, use Django database settings (default: False)

        Using with Django
            After django is loaded, simply construct with

                pg = PG(use_django=True)

            Tables can be shown with
                pg.table_names()

            Django queries can be run with one line
            df = PG(use_django=True).query('SELECT * FROM my_table').df
        """
        use_django = kwargs.get("use_django", False)
        if use_django:
            self.safe_import("django")
            from django.conf import settings

            db = settings.DATABASES["default"]
            kwargs = {"host": db["HOST"], "user": db["USER"], "password": db["PASSWORD"], "dbname": db["NAME"]}
        env_translator = {"host": "PGHOST", "user": "PGUSER", "password": "PGPASSWORD", "dbname": "PGDATABASE"}
        conn_kwargs = {key: kwargs.get(key, os.environ.get(env_translator[key], None)) for key in env_translator.keys()}
        bad_keys = []
        for key in conn_kwargs.keys():
            if conn_kwargs[key] is None:
                bad_keys.append(key)
        if bad_keys:
            raise ValueError(f"The following connections params not specified {bad_keys}")
        self._conn_kwargs = conn_kwargs

    @classmethod
    def queryset_to_sql(cls, queryset):
        """
        Transform a Django queryset into formatted SQL that can be used in pg-admin.

        Args:
            queryset: A Django QuerySet object

        Returns:
            str: Formatted SQL query string
        """
        from django.db import connection

        sqlparse = cls.safe_import("sqlparse")
        cls.safe_import("django")
        sql, sql_params = queryset.query.get_compiler(using=queryset.db).as_sql()
        with connection.cursor() as cur:
            query = cur.mogrify(sql, sql_params)
        query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query

    def schema_names(self):
        """
        Get a list of all schema names in the database.

        Returns:
            pandas.DataFrame: DataFrame containing schema names
        """
        pg = copy.deepcopy(self)
        return pg.query("SELECT nspname FROM pg_catalog.pg_namespace").to_dataframe()

    @functools.lru_cache()
    def table_names(self, schema_name="public"):
        """
        Get a list of table names in the specified schema.

        Args:
            schema_name (str): Name of the schema to query (default: "public")

        Returns:
            pandas.DataFrame: DataFrame containing table names, sorted alphabetically
        """
        pg = copy.deepcopy(self)
        df = pg.query(
            f"SELECT table_name FROM information_schema.tables WHERE table_schema='{schema_name}'"
        ).to_dataframe()
        df = df.sort_values(by="table_name")
        return df

    def query(self, sql) -> "PG":
        """
        Set the SQL query to be executed.

        Args:
            sql (str): SQL query string

        Returns:
            PG: Self instance for method chaining
        """
        self._sql = sql
        return self

    def run(self) -> "PG":
        """
        Execute the current SQL query and store results.

        Returns:
            PG: Self instance for method chaining

        Raises:
            psycopg2.ProgrammingError: If there's an error executing the query
        """
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:  # type: ignore
            with connection.cursor() as cursor:
                cursor.execute(self._sql)
                try:
                    self._raw_results = list(cursor.fetchall())
                    self._raw_columns = [col[0] for col in cursor.description]
                except psycopg2.ProgrammingError as e:
                    if str(e) == "no results to fetch":
                        self._raw_results = []
                        self._raw_columns = []
                    else:
                        raise
        return self

    def reset(self):
        """
        Reset the query results and columns to None.
        """
        self._raw_columns = None
        self._raw_results = None

    @property
    def _results(self) -> List[Tuple]:
        """
        Get the raw query results, executing the query if necessary.

        Returns:
            List[Tuple]: List of result tuples
        """
        if self._raw_results is None:
            self.run()
        return self._raw_results  # type: ignore

    @property
    def columns(self) -> List[str]:
        """
        Get the column names from the query results, executing the query if necessary.

        Returns:
            List[str]: List of column names
        """
        if self._raw_columns is None:
            self.run()
        return self._raw_columns  # type: ignore

    def as_tuples(self) -> List[Tuple]:
        """
        Get query results as a list of tuples.

        Returns:
            List[Tuple]: List of result tuples
        """
        self.reset()
        return self._results

    def as_dicts(self) -> List[Dict[str, Any]]:
        """
        Get query results as a list of dictionaries.

        Returns:
            List[Dict[str, Any]]: List of dictionaries with column names as keys
        """
        self.reset()
        return [dict(zip(self.columns, row)) for row in self._results]

    def as_named_tuples(self, named_tuple_name="Result") -> List[Any]:
        """
        Get query results as a list of named tuples.

        Args:
            named_tuple_name (str): Name for the named tuple class (default: "Result")

        Returns:
            List[Any]: List of named tuples
        """
        self.reset()
        nt_result = namedtuple(named_tuple_name, self.columns)
        return [nt_result(*row) for row in self._results]

    @classmethod
    def safe_import(cls, module_name, package=None):
        """
        Safely import a module, raising a helpful error if import fails.

        Args:
            module_name (str): Name of the module to import
            package (str, optional): Package name for relative imports

        Returns:
            module: The imported module

        Raises:
            ImportError: If the module cannot be imported
        """
        try:
            imported = importlib.import_module(module_name, package=package)
        except ImportError:
            raise ImportError(f"\n\nNope! This method requires that {module_name} be installed.  You know what to do.")
        return imported

    def as_dataframe(self) -> Any:
        """
        Get query results as a pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        self.reset()
        pd = self.safe_import("pandas")
        return pd.DataFrame(self._results, columns=self.columns)

    def to_tuples(self) -> List[Tuple]:
        """
        Alias for as_tuples().

        Returns:
            List[Tuple]: List of result tuples
        """
        return self.as_tuples()

    def to_dicts(self) -> List[Dict[str, Any]]:
        """
        Alias for as_dicts().

        Returns:
            List[Dict[str, Any]]: List of dictionaries with column names as keys
        """
        return self.as_dicts()

    def to_named_tuples(self) -> List[Any]:
        """
        Alias for as_named_tuples().

        Returns:
            List[Any]: List of named tuples
        """
        return self.as_named_tuples()

    def to_dataframe(self) -> Any:
        """
        Alias for as_dataframe().

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        return self.as_dataframe()

    @property
    def df(self):
        """
        Property alias for to_dataframe().

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        return self.to_dataframe()

    @property
    def sql(self) -> str:
        """
        Get the formatted SQL query string.

        Returns:
            str: Formatted SQL query string
        """
        sqlparse = self.safe_import("sqlparse")
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:  # type: ignore
            with connection.cursor() as cursor:
                sql, params = self._get_prepared_query()  # type: ignore
                query = cursor.mogrify(sql, params)
                query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query


def sql_file_to_df(file_name="sql_query.sql", context_dict=None):
    """
    Load and execute a SQL query from a file, returning the results as a DataFrame.
    Always connects to the postgres database with credentials from the environment.

    Args:
        file_name (str): Path to the SQL file. Defaults to 'sql_query.sql'.
        context_dict (dict, optional): Dictionary of variables to use for template rendering.
            Defaults to None.

    Returns:
        pandas.DataFrame: Results of the SQL query.
    """
    import jinja2

    if context_dict is None:
        context_dict = {}
    file_path = Path(file_name).expanduser().resolve()
    dir_path = file_path.parent
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(dir_path.as_posix()), autoescape=jinja2.select_autoescape()
    )
    template = jinja_env.get_template(file_path.name)
    text = template.render(**context_dict)
    pg = ezr.PG()
    pg.query(text)
    df = pg.to_dataframe()
    return df


def sql_string_to_df(query, context_dict=None):
    """
    Execute a SQL query string with Jinja templating, returning the results as a DataFrame.
    Always connects to the postgres database with credentials from the environment.

    Args:
        query (str): SQL query string with optional Jinja template variables.
        context_dict (dict, optional): Dictionary of variables to use for template rendering.
            Defaults to None.

    Returns:
        pandas.DataFrame: Results of the SQL query.
    """
    import jinja2

    if context_dict is None:
        context_dict = {}
    jinja_env = jinja2.Environment(loader=jinja2.BaseLoader(), autoescape=jinja2.select_autoescape())
    template = jinja_env.from_string(query)
    rendered_query = template.render(**context_dict)
    pg = ezr.PG()
    pg.query(rendered_query)
    df = pg.to_dataframe()
    return df



def _next_power_two(x):
    """given a number, returns the next power of two"""
    x = int(x)
    n = 1
    while n < x:
        n = n << 1
    return n


def _compute_pad(t, interp_exponent=0):
    """
    Given a sorted time series t, compute the zero padding.
    The final padded arrays are the next power of two in length multiplied
    by 2 ** interp_exponent.
    returns t_pad and y_pad
    """
    import numpy as np

    t_min, t_max, n = (t[0], t[-1], len(t))
    dt = (t_max - t_min) / float(n - 1)
    n_padded = _next_power_two(len(t)) << interp_exponent
    n_pad = n_padded - n
    t_pad = np.linspace(t_max + dt, t_max + dt + (n_pad - 1) * dt, n_pad)
    y_pad = np.zeros(len(t_pad))
    return (t_pad, y_pad)


def _compute_params(t):
    """
    Takes a timeseries and computes the parameters needed for the fast
    lomb scargle algorithm in gatspy
    """
    t_min, t_max, n = (t[0], t[-1], len(t))
    dt = (t_max - t_min) / float(n - 1)
    min_freq = 1.0 / (t_max - t_min)
    d_freq = 1.0 / (2 * dt * len(t))
    return (min_freq, d_freq, len(t))


def lomb_scargle(time, value, interp_exponent=0, freq_order=False):
    """Compute the Lomb-Scargle periodogram for a time series.

    This function performs a Lomb-Scargle periodogram analysis on a time series,
    which is useful for finding periodic signals in unevenly sampled data.

    Args:
        time (pandas.Series or array-like): Time values
        value (pandas.Series or array-like): Observed values
        interp_exponent (int, optional): Power of two for spectrum interpolation.
            Defaults to 0.
        freq_order (bool, optional): If True, return results in frequency order
            instead of period order. Defaults to False.

    Returns:
        pandas.DataFrame: A dataframe containing the periodogram results with columns:
            - period: The period of each frequency component
            - freq: The frequency of each component
            - power: The power spectrum
            - amp: The amplitude spectrum

    Note:
        The input time series is automatically:
        - Sorted by time
        - Mean-centered
        - Zero-padded to the next power of two
    """
    import numpy as np
    import gatspy
    import pandas as pd

    if not isinstance(time, pd.Series):
        time = pd.Series(time)
    if not isinstance(value, pd.Series):
        value = pd.Series(value)
    if len(time) != len(value):
        raise ValueError(
            f"Time and value arrays must have the same length. Got time length {len(time)} and value length {len(value)}"
        )
    df = pd.DataFrame({"t": time, "y": value}).dropna()
    df = df.sort_values(by=["t"])
    df["y"] = df["y"] - df.y.mean()
    E_in = np.sum(df.y * df.y)
    pre_pad_length = len(df)
    t_pad, y_pad = _compute_pad(df.t.values, interp_exponent=interp_exponent)
    if len(t_pad) > 0:
        df = pd.concat([df, pd.DataFrame({"t": t_pad, "y": y_pad})], ignore_index=True)
    model = gatspy.periodic.LombScargleFast()
    model.fit(df.t.values, df.y.values, 1)
    f0, df, N = _compute_params(df.t.values)
    f = f0 + df * np.arange(N)
    p = 1.0 / f
    yf = model.score_frequency_grid(f0, df, N)
    yf_power = 2 * yf * E_in * len(yf) / float(pre_pad_length) ** 2
    yf_amp = np.sqrt(yf_power)
    df = pd.DataFrame({"freq": f, "period": p, "power": yf_power, "amp": yf_amp})[["period", "freq", "power", "amp"]]
    if not freq_order:
        df = df.sort_values(by="period")  # type: ignore
    return df



import textwrap


class StreamPlotter:
    """
    Make jupyter streaming plots using matplotlib.
    You can track multiple overlaid plots on multiple
    subplots.  The entire matplotlib figure is redrawn
    for each call to .track(), so the rendering performanc
    is slow.  Requires the ipympl package.

    pip install ipympl

    Then, in the notebook:
    %matplotlib ipympl
    """

    example = textwrap.dedent(
        "\n        # Somewhere at the top of the notebook\n        %matplotlib ipympl\n\n        # In a jupyter cell you want to show the live plots\n        # (Don't need any kwargs if you only want a single plot)\n        plotter = ezr.StreamPlotter(rows=1, cols=2)\n        plotter.init()\n\n        # In a different jupyter cell with code you want to track\n        t = np.linspace(0, 2 * np.pi, 30)\n        for t0 in np.linspace(0, 2 * np.pi, 50):\n            ys = np.sin(t0 - t)\n            yc = np.cos(t0 - t)\n            plotter.update(t, ys, subplot=0, name='left sin', plotspec='b.-', )\n            plotter.update(t, yc, subplot=0, name='left cos', plotspec='r', )\n            plotter.update(t, yc, subplot=1, name='right ysp', plotspec='b', )\n            plotter.update(t, -yc, subplot=1, name='right ysm', plotspec='r', )\n    "
    )

    def __init__(self, rows=1, cols=1):
        """
        Args:
            rows: number of subplot rows
            cols: number of subplot cols
        """
        self.line_dict = {}
        self.rows = rows
        self.cols = cols

    def init(self):
        """
        Initializes a tracker by setting up the axes on which
        plots will be made.  This should be done in a different
        jupyter notebook cell than that in which the .track()
        method is called.
        """
        from matplotlib import pyplot as plt

        self.figure, self.ax_list = plt.subplots(self.rows, self.cols, figsize=(9, 5))
        if self.rows * self.cols == 1:
            self.ax_list = [self.ax_list]
        else:
            self.ax_list = self.ax_list.flatten()
        for ax in self.ax_list:
            ax.grid(True)
        plt.show(block=False)

    def update(self, x, y, name="metric", subplot=0, plotspec="k-", logy=False):
        """
        Call this method in your loop to update the plots.
        Args:
            x: The x values to plot
            y: The y values to plot
            name: A label to attach to this curve
            subplot: Which subplot to place the plot in
            plotspec: The matplotlib plot specification.
            logy: Set the y-axis to log scale

        """

        if not 0 <= subplot < len(self.ax_list):
            raise ValueError(f"Must have 0 <= subplot < {len(self.ax_list)}")
        ax = self.ax_list[subplot]
        if logy:
            ax.set_yscale("log")
        line = self.line_dict.get(name)
        if line is None:
            line = self.line_dict[name] = ax.plot(x, y, plotspec, label=name)[0]
            ax.legend(loc="best")
        else:
            line.set_data(x, y)
            ax.relim()
            ax.autoscale_view()
        self.figure.canvas.draw()



__version__ = "1.9.6"


from .hvtools import beta_plots, hist
from .bernstein import Bernstein, BernsteinFitter
from .utils import (
    BlobAttr,
    BlobMixin,
    cached_container,
    cached_dataframe,
    cached_property,
    diff_strings,
    get_logger,
    mute_warnings,
    pickle_cache_mixin,
    pickle_cache_state,
    pickle_cached_container,
    print_error,
    Scaler,
)
from .clock import Clock, GlobalClock
from .crypt import Crypt
from .distributions import DistFitter
from .duckcacher import DuckCacher, duckloader_factory
from .duckcacher import duckloader_factory as get_duckloader  # This is an alias
from .duck_mirror import DuckMirror
from .filtering.elliptic_filter import Elliptic
from .fit_lib import Fitter, classifier_evaluation_plots
from .gemini import Gemini
from .gsheet import GSheet as Gsheet
from .item import Item
from .metabase import Metabase
from .postgres import PG, pg_creds_from_env, sql_file_to_df, sql_string_to_df
from .param_state import ParamState
from .print_catcher import PrintCatcher
from .proportion_lib import Proportion
from .shaper import Shaper
from .timer import Timer
from .plotting import figure
from .ecdf_lib import ecdf
from .dataframe_tools import (
    column_level_flattener,
    events_from_starting_ending,
    heatmap,
    hex_from_dataframe,
    hex_from_duckdb,
    hex_to_dataframe,
    hex_to_duckdb,
    localize_utc_to_timezone,
    month_string,
    pandas_time_to_utc_seconds,
    pandas_utc_seconds_to_time,
    slugify,
    weekday_string,
)
from .iterify import iterify
from .outlier_tools import kill_outliers_iqr, kill_outliers_sigma_edit
from .lomb_scargle import lomb_scargle
from .memory import mem_get, mem_show
from .normal_sampler import NormalSampleJoiner
from .stream_plotter import StreamPlotter
from .tracker import Tracker
from .filtering.tvd import tvd
from .vonmises import VonMisesFitter
from .ez_agent import EZAgent
from .agent_runner import AgentRunner

from .dataframe_tools import get_quick_schema_class, get_pandas_sql_class
from .plotting import ColorCyle

QuickSchema = get_quick_schema_class()
PandasSql = get_pandas_sql_class()

cc = ColorCyle()

# Aliases for backwards compatibility
warnings_mute = mute_warnings

# Export list
__all__ = [
    "AgentRunner",
    "Bernstein",
    "BernsteinFitter",
    "BlobAttr",
    "BlobMixin",
    "Clock",
    "Crypt",
    "DistFitter",
    "DuckCacher",
    "DuckMirror",
    "duckloader_factory",
    "Elliptic",
    "EZAgent",
    "Fitter",
    "Gemini",
    "GlobalClock",
    "Gsheet",
    "Item",
    "Metabase",
    "PG",
    "ParamState",
    "PrintCatcher",
    "Proportion",
    "QuickSchema",
    "PandasSql",
    "Scaler",
    "Shaper",
    "Timer",
    "beta_plots",
    "cached_container",
    "cached_dataframe",
    "cached_property",
    "cc",
    "classifier_evaluation_plots",
    "column_level_flattener",
    "diff_strings",
    "ecdf",
    "events_from_starting_ending",
    "figure",
    "get_duckloader",
    "get_logger",
    "heatmap",
    "hex_from_dataframe",
    "hex_from_duckdb",
    "hex_to_dataframe",
    "hex_to_duckdb",
    "hist",
    "iterify",
    "kill_outliers_iqr",
    "kill_outliers_sigma_edit",
    "localize_utc_to_timezone",
    "lomb_scargle",
    "mem_get",
    "mem_show",
    "month_string",
    "mute_warnings",
    "NormalSampleJoiner",
    "pandas_time_to_utc_seconds",
    "pandas_utc_seconds_to_time",
    "pg_creds_from_env",
    "pickle_cache_mixin",
    "pickle_cache_state",
    "pickle_cached_container",
    "print_error",
    "slugify",
    "sql_file_to_df",
    "sql_string_to_df",
    "StreamPlotter",
    "Tracker",
    "tvd",
    "VonMisesFitter",
    "warnings_mute",
    "weekday_string",
]



from collections import OrderedDict
from itertools import chain
from textwrap import dedent
import copy


class examples:
    """
    A descriptor whose only purpose is to print help text
    """

    def __get__(self, *args, **kwargs):
        print(
            dedent(
                "\n                #=================================================================\n                #                Example Usage\n                #=================================================================\n\n                # Do imports\n                import numpy as np\n                from scipy.optimize import fmin\n                from easier import ParamState\n\n                # Define a model that gives response values in terms of params\n                def model(p):\n                    return p.a * p.x_train ** p.n\n\n                # Define a cost function for the optimizer to minimize\n                def cost(args, p):\n                    '''\n                    args: a numpy array of parameters that scipy optimizer passes in\n                    p: a ParamState object\n                    '''\n\n                    # Update paramstate with the latest values from the optimizer\n                    p.ingest(args)\n\n                    # Use the paramstate to generate a \"fit\" based on current params\n                    y_fit = model(p)\n\n                    # Compute the errors\n                    err = y_fit - p.y_train\n\n                    # Compute and return the cost\n                    cost = np.sum(err ** 2)\n                    return cost\n\n                # Make some fake data\n                x_train = np.linspace(0, 10, 100)\n                y_train = -7 * x_train ** 2\n                y_train = y_train + .5 * np.random.randn(len(x_train))\n\n\n                # Create a paramstate with variable names\n                p = ParamState('a n')\n\n                # Specify the data you are fitting\n                p.given(\n                    x_train=x_train,\n                    y_train=y_train\n                )\n\n\n                # Get the initial values for params\n                x0 = p.array\n\n                # Run the minimizer to get the optimal params\n                xf = fmin(cost, x0, args=(p,))\n\n                # Update ParamState with optimal params\n                p.ingest(xf)\n\n                # Print the optimized results\n                print(p)\n            "
            )
        )
        return None


class ParamState(object):
    """
    See the ParamState.examples attribute for examples
    """

    INITIAL_VALUE = 1.0
    examples = examples()

    def __init__(self, *args, **kwargs):
        """
        *args: a list of variable names
        **kwargs: any initializations to set initial arg values
        """
        from astropy.units.quantity import Quantity

        if len(args) == 1 and isinstance(args[0], str):
            args = args[0].replace(",", " ").split()
        self.vars = OrderedDict()
        self._fixed_vars = set()
        self.unit_dict = {}
        for arg in chain(args, kwargs.keys()):
            if arg not in self.vars:
                val = kwargs.get(arg, self.INITIAL_VALUE)
                try:
                    if isinstance(val, Quantity):
                        self.unit_dict[arg] = val.unit
                except NameError:
                    pass
                self.add(arg, val)

    def add(self, arg, val=None):
        if val is None:
            val = self.INITIAL_VALUE
        self.vars[arg] = val
        setattr(self, arg, val)

    def __setattr__(self, name, value):
        if hasattr(self, "vars") and name in self.vars:
            self.vars[name] = value
        super(ParamState, self).__setattr__(name, value)

    def clone(self):
        return copy.deepcopy(self)

    def drop(self, *variable_names):
        self._fixed_vars = self._fixed_vars - set(variable_names)
        for v in variable_names:
            if v in self.vars:
                del self.vars[v]
            if v in self.unit_dict:
                del self.unit_dict[v]

    def as_dict(self, copy=False):
        if copy:
            return OrderedDict(**self.vars)
        else:
            return self.vars

    def to_dict(self, copy=False):
        return self.as_dict(copy=copy)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def keys(self):
        return self.vars.keys()

    def values(self):
        return self.vars.values()

    def items(self):
        return self.vars.items()

    def ingest(self, array):
        """
        Update the parameter state with values obtained from array
        """
        variables = [k for k in self.vars.keys() if k not in self._fixed_vars]
        if len(variables) != len(array):
            raise ValueError("Array to ingest should have length {}".format(len(variables)))
        updates = dict(zip(variables, array))
        for var, val in updates.items():
            updates[var] = self.unit_dict.get(var, 1) * val
        self.vars.update(updates)
        self.__dict__.update(updates)

    def given(self, **kwargs):
        """
        Supply kwargs that will set constant variables.  These can either
        be new variables or already defined variables.  This allows you
        to easily turn on and off variables to optimize.
        """
        self.vars.update(kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._fixed_vars = self._fixed_vars.union(set(kwargs.keys()))
        return self

    @property
    def given_params(self):
        return OrderedDict([(k, v) for k, v in self.vars.items() if k in self._fixed_vars])

    @property
    def free_params(self):
        return OrderedDict([(k, v) for k, v in self.vars.items() if k not in self._fixed_vars])

    def __str__(self):
        """
        Prints a nice representation of the results
        """
        return self.df.to_string()

    def _repr_html_(self):
        return self.df.to_html()

    def __repr__(self):
        return self.__str__()

    @property
    def df(self):
        import pandas as pd

        rec_list = []
        for k, v in self.vars.items():
            kind = "*" if k in self._fixed_vars else ""
            rec_list.append((k, v, kind))
        df = pd.DataFrame(rec_list, columns=["var", "val", "const"])  # type: ignore
        df = df.sort_values(by=["const", "var"])
        df = df.set_index("var")
        df.index.name = None
        return df

    @property
    def array(self):
        """
        Returns only the non-constant variables as an array.  Use this for
        creating the initial state
        """
        from astropy.units.quantity import Quantity
        import numpy as np

        out = []
        for key, val in self.vars.items():
            if key not in self._fixed_vars:
                try:
                    if isinstance(val, Quantity):
                        val = val.value
                except NameError:
                    pass
                out.append(val)
        return np.array(out)

    @property
    def tuple(self):
        """
        Return a tuple of all variables.  Use this for calling functions
        """
        return tuple(self.vars.values())



from copy import deepcopy


class Item(object):
    """
    Item is a simple container class that sets its attributes from constructor
    kwargs.  It supports both object and dictionary access to its attributes.
    So, for example, all of the following statements are supported.

    If the value associated with a key is a dict, that value will recursively
    be transformed into an Item.

    .. code-block:: python

       item = Item(a=1, b=2)
       item['c'] = 2
       a = item['a']
       item_dict = item.as_dict()
    """

    # I'm using unconventional "_item_self_" name here to avoid
    # conflicts when kwargs actually contain a "self" arg.

    def __init__(_item_self, **kwargs):
        for key, val in kwargs.items():
            # Recursively create items for dict members
            if isinstance(val, dict):
                val = Item(**val)
            _item_self[key] = val

    def __str__(_item_self):
        quoted_keys = ["'{}'".format(k) for k in sorted(vars(_item_self).keys())]
        att_string = ", ".join(quoted_keys)
        return "Item({})".format(att_string)

    def __repr__(_item_self):
        return _item_self.__str__()

    def __setitem__(_item_self, key, value):
        setattr(_item_self, key, value)

    def __getitem__(_item_self, key):
        return getattr(_item_self, key)

    def __len__(self):
        return len(self.__dict__)

    def as_dict(self, copy=False):
        if copy:
            return dict(**self.__dict__)
        else:
            return self.__dict__

    def to_dict(self, copy=False):
        return self.as_dict(copy=copy)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def clone(self):
        return deepcopy(self)



class iterify:
    """
    This class wraps arguments that can either be iterable or not.
    If an iterable variable is provided, then that iterable variable
    is provided back from the .get_input_value() method and can be used
    in the function.

    If a non-iterable variable is provided, that variable is transformed
    to a single-element iterable and returned from the get_input_value() method.

    The get_output_value() method will return an object with the iterable property
    match what the input object had.

    This is useful if you want a function to handle both iterable and non-iterable
    versions of an argument.  You can just write the iterable code and this wrapper
    will trasform non iterable inputs to iterable, and then tranform outputs back
    to non-iterable.
    """

    def __init__(self, iter_constructor=list):
        self.iter_constructor = iter_constructor

    def _update_constructor(self, obj):
        import numpy as np

        if isinstance(obj, np.ndarray) and self.iter_constructor is list:
            self.iter_constructor = np.array

    def _make_iterable(self, x):
        if hasattr(x, "__iter__"):
            if self._constructed_type == type(x):
                return x
            else:
                return self.iter_constructor(x)
        else:
            return self.iter_constructor([x])

    def _store_state(self, obj):
        self._is_iterable = hasattr(obj, "__iter__")
        self._input_type = type(obj)
        if self._is_iterable:
            self._constructed_type = type(self.iter_constructor(obj[:2]))
        else:
            self._constructed_type = list

    def get_input(self, obj):
        self._update_constructor(obj)
        self._store_state(obj)
        return self._make_iterable(obj)
        if self.is_iterable:
            return obj
        else:
            return self.iter_constructor([obj])

    def get_output(self, obj):
        if self._is_iterable:
            if not isinstance(obj, self._constructed_type):
                obj = self.iter_constructor(obj)
            return obj
        else:
            return obj[0]



# import contextlib
# import dataset
# import os
# import pandas as pd
# import textwrap

# import os
# import textwrap
# import contextlib

# @contextlib.contextmanager
# def dataset_connection(file_name):
#     conn = dataset.connect(f'sqlite:///{file_name}')
#     try:
#         yield conn
#     finally:
#         conn.close()

# class Tables:
#     """
#     This descriptor provides attribute acces to the tables in the db.
#     Think of it like the .objects attribute on a Django model
#     """

#     def __init__(self, **kwargs):
#         self._table_names = []
#         for k, v in kwargs.items():
#             self._table_names.append(k)
#             setattr(self, k, v)

#     def __str__(self):
#         return f'Tables({self._table_names})'

#     def __repr__(self):
#         return self.__str__()

# class MiniTable:
#     """
#     A wrapper around a dataset table that knows how to dump to
#     a pandas dataframe
#     """

#     def __init__(self, file_name, table_name):
#         self._table_name = table_name
#         self._file_name = file_name

#     def __str__(self):
#         return f'MiniTable({self._table_name})'

#     def __repr__(self):
#         return self.__str__()

#     @property
#     def df(self):
#         with dataset_connection(self._file_name) as connection:
#             table = connection[self._table_name]
#             df = pd.DataFrame(table.all())
#         return df

# class table_getter:

#     def __get__(self, obj, cls):
#         content = {}
#         for table_name in obj.table_names:
#             content[table_name] = MiniTable(obj.file_name, table_name)
#         return Tables(**content)

# class MiniModel:
#     """
#     A class for storing pandas dataframes in sqlite database
#     """
#     tables = table_getter()
#     example = textwrap.dedent("\n        import pandas as pd\n        import easier as ezr\n\n        # Creat dataframe to store\n        df_one = pd.DataFrame(\n            [\n                {'a': 1, 'b': pd.Timestamp('1/1/2022')},\n                {'a': 2, 'b': pd.Timestamp('1/2/2022')}\n            ]\n        )\n\n        # Create second dataframe to store\n        df_two = df_one.copy()\n        df_two.loc[0, 'a'] = 7\n        df_two.loc[0, 'b'] = pd.Timestamp('11/18/2022')\n\n\n        # Create a sqlite file managed by minimodel\n        mm = ezr.MiniModel('rob.sqlite', overwrite=True)\n\n        # Create a table from the first dataframe\n        mm.create('table_one', df_one)\n\n        # Insert additional records to the first table\n        mm.insert('table_one', df_two)\n\n        # Create a second table from the second dataframe\n        # (Note I'm creating the table using insert, which\n        # is also allowed)\n        mm.insert('table_two', df_two)\n\n        # Get the saved table as a dataframe\n        df1 = mm.tables.table_one.df\n        df2 = mm.tables.table_two.df\n        print()\n        print(df1.to_string())\n        print()\n        print(df2.to_string())\n\n        # Create a third dataframe to demonstrate upserting\n        dfup = df1.copy().drop('id', axis=1)\n        dfup.loc[:, 'b'] = pd.Timestamp('11/18/2050')\n\n        # Upsert into table two\n        df3_pre = mm.tables.table_two.df\n        mm.upsert('table_two', ['a'], dfup)\n        df3_post = mm.tables.table_two.df\n        print()\n        print(df3_pre.to_string())\n        print()\n        print(df3_post.to_string())\n    ")

#     def __init__(self, file_name='./mini_model.sqlite', overwrite=False, read_only=False):
#         """
#         Args:
#             file_name: The name of the sqlite file
#             overwrite: Blow away any existing file with that name and ovewrite
#             read_only: Prevents unintentional overwriting of a database
#         """
#         self.file_name = os.path.realpath(os.path.expanduser(file_name))
#         self._read_only = read_only
#         if overwrite and os.path.isfile(file_name):
#             os.unlink(file_name)

#     def __str__(self):
#         return f'MiniModel({os.path.basename(self.file_name)})'

#     def __repr__(self):
#         return self.__str__()

#     @property
#     def table_names(self):
#         """
#         A utility property that returns all the table names in the db
#         """
#         names = []
#         if os.path.isfile(self.file_name):
#             with dataset_connection(self.file_name) as connection:
#                 names = connection.inspect.get_table_names()
#         return names

#     def _framify(self, data):
#         """
#         A utility method to turn inputs into dataframe.
#         Inputs can be dicts, series or dataframes.

#         If dicts or series, they are transformed into rows of a dataframe
#         """
#         if isinstance(data, dict):
#             data = pd.Series(data)
#         if isinstance(data, pd.Series):
#             data = pd.DataFrame(data).T
#         if not isinstance(data, pd.DataFrame):
#             raise ValueError('data must be a dict, pandas series or pandas dataframe')
#         return data

#     def _listify(self, df):
#         """
#         Utility method to turn a dataframe into a list of dicts that dataset
#         knows how to insert into the db
#         """
#         datetime_cols = [c for c in df.columns if isinstance(df[c].iloc[0], pd.Timestamp)]
#         recs = df.to_dict('records')
#         for col in datetime_cols:
#             for rec in recs:
#                 rec[col] = rec[col].to_pydatetime()
#         return recs

#     def _ensure_writeable(self):
#         if self._read_only:
#             raise ValueError('Trying to write to read-only db')

#     def insert(self, table_name, data):
#         """
#         Method for inserting data (series, dict or dataframe) into db
#         """
#         self._ensure_writeable()
#         df = self._framify(data)
#         recs = self._listify(df)
#         with dataset_connection(self.file_name) as connection:
#             connection[table_name].insert_many(recs)
#         return self

#     def create(self, table_name, data):
#         """
#         Creates (i.e. overwrites) a table in the database with the supplied data
#         """
#         self._ensure_writeable()
#         if table_name in self.table_names:
#             with dataset_connection(self.file_name) as connection:
#                 connection[table_name].drop()
#         self.insert(table_name, data)
#         return self

#     def upsert(self, table_name, keys, data):
#         """
#         Upsert records into specified table.
#         Args:
#             table_name: the name of the table to upsert
#                   keys: a list of column names that the input data must match
#                         in the db for a record to be updated in stead of inserted
#                   data: The data to upsert.  Single records can be series or dicts.
#                         Multiple records defined using dataframes

#         """
#         self._ensure_writeable()
#         df = self._framify(data)
#         recs = self._listify(df)
#         with dataset_connection(self.file_name) as connection:
#             connection[table_name].upsert_many(recs, keys)
#         return self

#     def query(self, sql):
#         """
#         Run a SQL query against the database
#         """
#         with dataset_connection(self.file_name) as connection:
#             df = pd.DataFrame(connection.query(sql))
#         return df



from ..bernstein import Bernstein, Compress, BernsteinFitter
import numpy as np
import pytest

from ..bernstein import Bernstein, Compress, BernsteinFitter
import pytest

def get_xy():
    x = np.linspace(0, np.pi / 4, 300)
    y = np.sin(x)
    dydx = np.cos(x)
    return (x, y, dydx)

def test_bernstein_simple():
    x, y, dydx = get_xy()
    yf = Bernstein(x=x, y=y, N=1000).predict(x)
    dydxf = Bernstein(x=x, y=y, N=1000).predict_derivative(x)
    max_delta = np.max(np.abs(yf - y))
    max_delta_deriv = np.max(np.abs(dydxf - dydx))
    assert max_delta < 0.0001
    assert max_delta_deriv < 0.001

def test_bernstein_limits():
    x, y, dydx = get_xy()
    yf = Bernstein(x=x, y=y, N=1000, xlim=(0, np.pi / 4)).predict(x)
    max_delta = np.max(np.abs(yf - y))
    assert max_delta < 0.0001

def test_bernstein_bad_compress_type():
    with pytest.raises(ValueError):
        Compress().compress([1, 2, 3])

def test_bernstein_compress_dont_learn_lims():
    cmp = Compress(min_val=0, max_val=np.pi)
    xc = cmp.compress(np.array([1, 2, 3]), learn_limits=False)
    assert tuple([int(v) for v in cmp.expand(xc)]) == (1, 2, 3)

def test_bernstein_compress_dont_learn_bad_limits():
    cmp = Compress()
    with pytest.raises(ValueError):
        cmp.compress(np.array([1, 2, 3]), learn_limits=False)

def test_bernstein_too_short():
    x = np.array([1, 2])
    with pytest.raises(ValueError):
        Bernstein(x=x, y=x, N=1000).predict(x)

def test_bernstein_not_numpy():
    x = [1, 2, 3]
    with pytest.raises(ValueError):
        Bernstein(x=x, y=x, N=1000).predict(x)

def test_bernstein_fitter():
    x = np.linspace(0, 7 * np.pi / 3, 300)
    y = -1 + np.sin(x) + 0.8 * x
    y[-1] = y[-1] + 1
    blob = BernsteinFitter(non_negative=False, monotonic=False, match_left=False, match_right=False).fit(x, y, 150).to_blob()
    b = BernsteinFitter().from_blob(blob)
    yf = b.predict(x)
    assert np.abs(np.max(y - yf)) < 0.1
    yf = BernsteinFitter(non_negative=False, monotonic=False, match_left=False, match_right=False).fit_predict(x, y, 5)
    assert np.abs(yf[0] - y[0]) > 1e-06
    assert np.abs(yf[-1] - y[-1]) > 1e-06
    assert np.min(yf) < 0
    assert np.min(np.diff(yf)) < 0
    yf = BernsteinFitter(non_negative=False, monotonic=False, match_left=True, match_right=True).fit_predict(x, y, 5)
    assert np.abs(yf[0] - y[0]) < 1e-06
    assert np.abs(yf[-1] - y[-1]) < 1e-06
    yf = BernsteinFitter(non_negative=True, monotonic=False, match_left=False, match_right=False).fit_predict(x, y, 5)
    assert np.min(y) < 0
    assert np.min(yf) > 0
    yf = BernsteinFitter(non_negative=False, monotonic=True, match_left=False, match_right=False).fit_predict(x, y, 15)
    assert np.min(np.diff(y)) < 0
    assert np.min(np.diff(yf)) >= -1e-05
    t = np.linspace(0, 2 * np.pi, 1000)
    ys = np.exp(np.sin(2 * t))
    dydt = np.diff(ys) / np.diff(t)
    fitter = BernsteinFitter(non_negative=False, monotonic=False).fit(t, ys, 55)
    dydtf = fitter.predict_derivative(t)[1:]
    assert np.max(np.abs(dydtf - dydt)) < 0.05


from functools import partial

from textwrap import dedent


class Proportion:
    """
    A class for analyzing proportions.

    ***************************************************
    print(ezr.Proportion.example)
    ***************************************************

    ...
    """

    example = dedent(
        "\n        import easier as ezr\n        # Create a \"small statistics\" proportion with win = 60%\n        ps = ezr.Proportion(3, 5, name='small')\n\n        # Create a \"large statistics proportion\n        # that is \"worse\" than the small one with win = 50%\n        pb = ezr.Proportion(500, 1000, name='big')\n\n        # We can set how to compare the distibutions\n        # Setting this to None (the default) will use\n        # the distribution means for comparison.  Setting\n        # it to some value in range (0, 1) will compare\n        # that percentile.\n        ps.set_default_global_comparison_quantile(.05)\n\n        # Print how much better the mean of one proportion is than the other\n        ps.set_default_global_comparison_quantile(None)\n        print(f'Big is {pb - ps:0.2f} better than small based on mean')\n\n        # Print how much better a quantile of one dist is than the other\n        ps.set_default_global_comparison_quantile(.05)\n        print(f'Big is {pb - ps:0.2f} better than small based on q=.05')\n\n        # All comparison operators are defined\n        print('big<small', pb < ps)\n        print('big<=small', pb <= ps)\n        print()\n        print('big>small', pb > ps)\n        print('big>=small', pb >= ps)\n        print()\n        print('big==small', pb == ps)\n        print(sorted([ps, pb]))\n\n        # I can plot the distributions on top of each other\n        display(ps.plot() * pb.plot())\n\n        # I can plot the probability that one proportion is\n        # greater than the other\n        display(pb.plot_prob_better_curve(ps))\n    "
    )
    _DEFAULT_COMPARISON_QUANTILE = None

    def __init__(self, num_won, num_total, name="", plot_range=(0, 1), num_plot_points=1000):
        """
        Args:
            num_won: The number of events "won"
            num_total: The total number of events
            plot_range: The range of the plot
            num_plot_points: The number of points to plot
        """
        from scipy import stats

        if num_won > num_total:
            raise ValueError("You cant win more than the total")
        if num_total < 1:
            raise ValueError("You must have at least one observation")
        self.num_won = num_won
        self.num_total = num_total
        self.alpha = num_won + 1
        self.beta = num_total - num_won + 1
        self.proportion = num_won / num_total
        self.dist = stats.beta(self.alpha, self.beta)
        self.name = name
        self.plot_range = plot_range
        self.num_plot_points = num_plot_points

    @property
    def comparison_quantile(self):
        return self._DEFAULT_COMPARISON_QUANTILE

    @property
    def nominal_value(self):
        if self.comparison_quantile is None:
            return self.dist.mean()
        else:
            return self.dist.ppf(self.comparison_quantile)

    def set_default_global_comparison_quantile(self, value):
        self.__class__._DEFAULT_COMPARISON_QUANTILE = value

    def _ensure_comparable(self, other):
        if len({self.comparison_quantile, other.comparison_quantile}) > 1:
            raise ValueError("Both proportion objects must have same comparsion_quantile")

    def __sub__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value - other.nominal_value

    def __eq__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value == other.nominal_value

    def __lt__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value < other.nominal_value

    def __le__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value <= other.nominal_value

    def __gt__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value > other.nominal_value

    def __ge__(self, other):
        self._ensure_comparable(other)
        return self.nominal_value >= other.nominal_value

    def __str__(self):
        if self.name == "":
            prefix = ""
        else:
            prefix = f"{self.name}="
        return f"P({prefix}{self.num_won}/{self.num_total})"

    def __repr__(self):
        return self.__str__()

    def probability_better_than(self, other, amount=0):
        from scipy import integrate

        my_dist = self.dist
        their_dist = other.dist

        def prob_better(x):
            prob_me_at_x = my_dist.pdf(x)  # type: ignore
            prob_they_are_less = their_dist.cdf(x - amount)
            prob_me_at_x_and_better_than_them = prob_me_at_x * prob_they_are_less
            return prob_me_at_x_and_better_than_them

        v_num = integrate.quad(prob_better, 0, 1, points=[self.proportion])[0]
        return v_num

    def plot(self):
        import numpy as np

        args = list(self.plot_range) + [self.num_plot_points]
        x = np.linspace(*args)
        c = hv.Curve((x, self.dist.pdf(x)), "Proportion", "Density", label=self.name)  # type: ignore
        return c

    def plot_prob_better_curve(self, other):
        import numpy as np
        from scipy import integrate
        import holoviews as hv

        my_dist = self.dist
        their_dist = other.dist

        def integrand(delta, x):
            prob_me_at_x = my_dist.pdf(x)  # type: ignore
            prob_at_least_delta_better = their_dist.cdf(x - delta)
            return prob_me_at_x * prob_at_least_delta_better

        def prob_at_least_better_than(delta):
            v_num = integrate.quad(partial(integrand, delta), 0, 1, points=[self.proportion, other.proportion])[0]
            return v_num

        my_min = self.dist.ppf(0.01)
        my_max = self.dist.ppf(0.99)
        their_min = other.dist.ppf(0.01)
        their_max = other.dist.ppf(0.99)
        minval = min([my_min, their_min])
        maxval = max([my_max, their_max])
        delta = maxval - minval
        if self.name == "":
            my_name = "current distribution"
        else:
            my_name = self.name
        if other.name == "":
            their_name = "other distribution"
        else:
            their_name = other.name
        xvals = np.linspace(-delta, delta, self.num_plot_points)
        yvals = [prob_at_least_better_than(x) for x in xvals]
        return hv.Curve(
            (xvals, yvals), f"{my_name!r} is at least this much better than {their_name!r}", "probability this is true"
        )



from typing import Union, Optional


class EZAgent:
    """
    A wrapper class for creating and configuring an Agent with simplified initialization.

    This class provides an easy way to create an agent with predefined safety settings
    and system prompts.
    """

    def __init__(
        self,
        system_prompt: str,
        model_name: str = "google-vertex:gemini-2.5-flash",
        retries: int = 1,
        validate_model_name: bool = True,
    ) -> None:
        """
        Initialize an EZAgent with the specified system prompt, model, and retry settings.

        Args:
            system_prompt: The system prompt to use for the agent.
            model_name: The name of the model to use for the agent. Defaults to 'google-vertex:gemini-2.5-pro'.
            retries: The number of retries to attempt if agent calls fail. Defaults to 1.
            validate_model_name: Whether to validate the model name against allowed models. Defaults to True.

        Raises:
            ValueError: If the model_name is not one of the allowed models and validate_model_name is True.
        """
        allowed_models = [
            "google-vertex:gemini-2.5-flash",
            "google-vertex:gemini-2.5-pro",
        ]

        if validate_model_name and model_name not in allowed_models:
            raise ValueError(f"Model {model_name} not supported. Please use one of: {', '.join(allowed_models)}")

        # Store model name as public attribute
        self.model_name = model_name

        from pydantic_ai import Agent
        from pydantic_ai.models.gemini import GeminiModelSettings

        self._GeminiModelSettings = GeminiModelSettings

        # Store model configuration as attributes instead of creating instance
        self._temperature = 0
        self._gemini_safety_settings = [
            {"category": "HARM_CATEGORY_HATE_SPEECH", "threshold": "OFF"},
            {"category": "HARM_CATEGORY_DANGEROUS_CONTENT", "threshold": "OFF"},
            {"category": "HARM_CATEGORY_SEXUALLY_EXPLICIT", "threshold": "OFF"},
            {"category": "HARM_CATEGORY_HARASSMENT", "threshold": "OFF"},
        ]

        self.agent = Agent(
            model_name,
            instructions=system_prompt,
            retries=retries,
        )

    async def run(
        self,
        user_prompt: str,
        output_type: Optional["pydantic.main.BaseModel"] = None,  # type: ignore
    ) -> Union["pydantic_ai.agent.AgentRunResult", None]:  # type: ignore
        """
        Run the agent with the given user prompt and expected output type.

        Args:
            user_prompt: The prompt to send to the agent.
            output_type: Optional Pydantic model for structured output.

        Returns:
            The agent run result, or None if the run fails.
        """
        # Create model settings with all stored configuration
        model_settings = self._GeminiModelSettings(
            temperature=self._temperature,
            gemini_safety_settings=self._gemini_safety_settings.copy(),  # type: ignore
        )

        result = await self.agent.run(
            user_prompt,
            output_type=output_type,
            model_settings=model_settings,
        )  # type: ignore
        return result

    def get_usage(self, *results: "pydantic_ai.agent.AgentRunResult") -> "pd.Series":  # type: ignore
        """
        Extract usage information from one or more agent run results and return as a pandas Series.

        Args:
            *results: One or more result objects returned from the agent's run method.

        Returns:
            A pandas Series containing aggregated token usage statistics with keys:
            - 'requests': Total number of API requests made
            - 'request_tokens': Total number of tokens in the input/prompts
            - 'response_tokens': Total number of tokens in the model's responses
            - 'thoughts_tokens': Total number of tokens used for internal reasoning
            - 'total_tokens': Total number of tokens used (request + response + thoughts)

        Raises:
            AttributeError: If any result object doesn't have a usage() method.
            ValueError: If no results are provided.
        """
        if not results:
            raise ValueError("At least one result must be provided.")

        try:
            import pandas as pd

            # Initialize totals
            total_requests = 0
            total_request_tokens = 0
            total_response_tokens = 0
            total_thoughts_tokens = 0
            total_tokens = 0

            # Aggregate usage from all results
            for result in results:
                usage = result.usage()
                total_requests += usage.requests
                total_request_tokens += usage.request_tokens
                total_response_tokens += usage.response_tokens
                total_tokens += usage.total_tokens

                # Extract thoughts_tokens from details if available
                if hasattr(usage, "details") and usage.details:
                    total_thoughts_tokens += usage.details.get("thoughts_tokens", 0)

            data = {
                "requests": total_requests,
                "request_tokens": total_request_tokens,
                "response_tokens": total_response_tokens,
                "thoughts_tokens": total_thoughts_tokens,
                "total_tokens": total_tokens,
            }
            return pd.Series(data)
        except AttributeError:
            raise AttributeError(
                "One or more result objects do not have a usage() method. Make sure you're passing valid agent run results."
            )



import easier as ezr


class DistFitter:

    def __init__(self, dist_class, num_hist_bins=None, hist_range=None):
        """
        Args:
            dist_class: A scipy.stats distribution class to fit to
            num_hist_bins: The number of bins to use for the empirical fitter
            hist_range: The range to use in the impirical fitter
        """
        self.dist_class = dist_class
        self._num_hist_bins = num_hist_bins
        self._hist_range = hist_range

    def fit(self, data, **fit_kwargs):
        """
        Args:
            data: iterable of data you want to fit to
        Kwargs:
            shape: The initial guess for the shape parameter(s) you want to pass to the fitter
            f0, f1, ...: Pin the shape parameter(s) to these values
            loc: The starting guess for the location parameter
            scale: the starting guess for the scale parameter
            floc: pin the lcoation parameter to this value
            fscale: pin the scale parameter to this value
        """
        from scipy import stats
        import numpy as np

        data = np.array(data)
        shape = fit_kwargs.pop("shape", None)
        if shape:
            args = (data, shape)
        else:
            args = (data,)
        self.dist_params = self.dist_class.fit(*args, **fit_kwargs)
        self.dist = self.dist_class(*self.dist_params)
        if self._num_hist_bins is None:
            hist_bins = int(len(data) / 5)
        else:
            hist_bins = self._num_hist_bins
        self.hist_dist = stats.rv_histogram(np.histogram(data, bins=hist_bins, range=self._hist_range))
        min_val, max_val = self.hist_dist.ppf([0.01, 0.99])
        if self._hist_range:
            min_val, max_val = self._hist_range
        self.x = np.linspace(min_val, max_val, 600)
        (self.q_theory, self.q_data), (self._a, self._b, _) = stats.probplot(data, dist=self.dist, plot=None, fit=True)
        return self

    def _plot(self, label, method_name, logx=False, logy=False, data_color=None, fit_color=None, xlabel="value"):
        import holoviews as hv

        if data_color is None:
            data_color = ezr.cc[0]
        if fit_color is None:
            fit_color = ezr.cc[1]
        func = getattr(self.dist, method_name)
        hist_func = getattr(self.hist_dist, method_name)
        c1 = hv.Curve((self.x, 1e-09 + func(self.x)), xlabel, "Density", label=f"{self.dist.dist.name} Fit").options(logx=logx, logy=logy, color=fit_color)  # type: ignore
        c2 = hv.Curve((self.x, 1e-09 + hist_func(self.x)), label=f"{label} Empirical").options(color=data_color)  # type: ignore
        return hv.Overlay([c1, c2]).options(legend_position="top")

    def plot_pdf(self, label="", logx=False, logy=False, data_color=None, fit_color=None, xlabel="value"):
        """
        Plot the pdf
        """
        return self._plot(
            label, method_name="pdf", logx=logx, logy=logy, data_color=data_color, fit_color=fit_color, xlabel=xlabel
        )

    def plot_cdf(self, label="", logx=False, logy=False, data_color=None, fit_color=None, xlabel="value"):
        """
        Plot the cdf
        """
        return self._plot(
            label, method_name="cdf", logx=logx, logy=logy, data_color=data_color, fit_color=fit_color, xlabel=xlabel
        )

    def plot_qq(self, label="", logx=False, logy=False, data_color=None, fit_color=None):
        """
        Make a q-q plot.  X axis will hold theoretical quantiles, Y axis the impirircal
        """
        import numpy as np
        import holoviews as hv

        if data_color is None:
            data_color = "blue"
        if fit_color is None:
            fit_color = "red"
        q_data_fit = np.polyval([self._a, self._b], self.q_theory)  # type: ignore
        c1 = hv.Scatter((self.q_theory, self.q_data), "Theoretical Quantiles", "Empirical Quantiles", label=label).options(size=5, alpha=0.5, color=data_color, logx=logx, logy=logy)  # type: ignore
        label = f"{self.dist.dist.name}  params = {['%0.4g' % p for p in self.dist_params]}".replace("'", "")
        c2 = hv.Curve((self.q_theory, q_data_fit), label="Best Fit Line").options(color=fit_color, alpha=0.2)  # type: ignore
        c3 = hv.Curve((self.q_theory, self.q_theory), label=label).options(color="green", alpha=0.2)  # type: ignore
        return hv.Overlay([c1, c2, c3]).options(legend_position="top")



import pandas as pd
from duckmirror import Duckmirror
from unittest.mock import patch
import tempfile
import os
import shutil
import pytest


def generate_test_df(df_type: str) -> pd.DataFrame:
    """Generate a test dataframe based on the specified type.

    Args:
        df_type (str): Type of dataframe to generate. Options:
            - 'numbers': Simple numeric dataframe
            - 'text': Simple text dataframe
            - 'mixed': Mixed numeric and text dataframe
            - 'empty': Empty dataframe with columns
            - 'dates': Dataframe with dates

    Returns:
        pd.DataFrame: Generated test dataframe
    """
    if df_type == "numbers":
        return pd.DataFrame({"id": [1, 2, 3], "value": [10.5, 20.5, 30.5]})
    elif df_type == "text":
        return pd.DataFrame(
            {"name": ["Alice", "Bob", "Charlie"], "category": ["A", "B", "A"]}
        )
    elif df_type == "mixed":
        return pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "score": [95.5, 88.0, 92.5],
            }
        )
    elif df_type == "empty":
        return pd.DataFrame(columns=["id", "name", "value"])
    elif df_type == "dates":
        return pd.DataFrame(
            {"date": pd.date_range("2024-01-01", periods=3), "value": [100, 200, 300]}
        )
    else:
        raise ValueError(f"Unknown dataframe type: {df_type}")


def test_register_decorator():
    """Test that the register decorator correctly stores and returns dataframes."""
    # Create a new Duckmirror instance with in-memory database
    mirror = Duckmirror(":memory:")

    # Create a test function that returns a mixed dataframe
    @mirror.register
    def test_table() -> pd.DataFrame:
        return generate_test_df("mixed")

    # Get the dataframe through the mirror
    result_df = mirror.df_test_table

    # Verify the dataframe contents
    expected_df = generate_test_df("mixed")
    pd.testing.assert_frame_equal(result_df, expected_df)

    # Verify we can access the same dataframe multiple times
    result_df2 = mirror.df_test_table
    pd.testing.assert_frame_equal(result_df2, expected_df)


def test_caching_behavior():
    """Test that the caching mechanism works correctly and returns copies."""
    # Create a new Duckmirror instance with in-memory database
    mirror = Duckmirror(":memory:")

    # Track function calls
    call_count = 0

    @mirror.register
    def cached_table() -> pd.DataFrame:
        nonlocal call_count
        call_count += 1
        return generate_test_df("numbers")

    # First call should execute the function
    df1 = mirror.df_cached_table
    assert call_count == 1, "Function should be called on first access"

    # Second call should use cache
    df2 = mirror.df_cached_table
    assert call_count == 1, "Function should not be called again"

    # Verify we get copies of the dataframe
    assert df1 is not df2, "Should get different dataframe objects"

    # Verify the dataframes are equal
    pd.testing.assert_frame_equal(df1, df2)

    # Verify modifying one dataframe doesn't affect the other
    df1.loc[0, "value"] = 999.9
    assert (
        df1.loc[0, "value"] != df2.loc[0, "value"]
    ), "Modifying one dataframe should not affect the other"


def test_cache_contents():
    """Test that the _cache attribute contains a copy of the dataframe only when cache_in_memory=True."""
    # Test with cache_in_memory=True
    mirror = Duckmirror(":memory:", cache_in_memory=True)

    @mirror.register
    def test_table() -> pd.DataFrame:
        return generate_test_df("numbers")

    # First access should populate cache
    df1 = mirror.df_test_table
    expected_df = generate_test_df("numbers")

    # Verify cache contains a copy of the dataframe
    cache_key = "main.test_table"
    assert cache_key in mirror._cache, "Cache should contain the dataframe"
    pd.testing.assert_frame_equal(mirror._cache[cache_key], expected_df)
    assert (
        mirror._cache[cache_key] is not df1
    ), "Cache should contain a copy, not the same object"

    # Test with cache_in_memory=False
    mirror = Duckmirror(":memory:", cache_in_memory=False)

    @mirror.register
    def test_table2() -> pd.DataFrame:
        return generate_test_df("numbers")

    # Access should not populate cache
    df2 = mirror.df_test_table2

    # Verify cache is empty
    assert len(mirror._cache) == 0, "Cache should be empty when cache_in_memory=False"


def test_run_query_call_count():
    # cache_in_memory=True: run_query should be called only once for multiple accesses
    mirror = Duckmirror(":memory:", cache_in_memory=True)
    with patch.object(
        mirror, "run_query_df", wraps=mirror.run_query_df
    ) as mock_run_query:

        @mirror.register
        def test_table() -> pd.DataFrame:
            return generate_test_df("numbers")

        # First access triggers run_query for CREATE TABLE and SELECT
        df1 = mirror.df_test_table
        expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df1, expected_df)

        # Verify no queries have been called yet
        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

        df2 = mirror.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)
        assert df1 is not df2, "Dataframes should be different objects (copies)"
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

    mirror = Duckmirror(":memory:", cache_in_memory=False)
    with patch.object(
        mirror, "run_query_df", wraps=mirror.run_query_df
    ) as mock_run_query:

        @mirror.register
        def test_table() -> pd.DataFrame:
            return generate_test_df("numbers")

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        print(select_queries)
        assert (
            len(select_queries) == 0
        ), "No SELECT query should have been executed for test_table"

        # First access triggers run_query for CREATE TABLE and SELECT
        df1 = mirror.df_test_table
        expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df1, expected_df)

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 1
        ), "One SELECT query should have been executed for test_table"

        df2 = mirror.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)
        assert df1 is not df2, "Dataframes should be different objects (copies)"

        queries = [call[0][0] for call in mock_run_query.call_args_list]
        select_queries = [q for q in queries if "SELECT * FROM main.test_table" in q]
        assert (
            len(select_queries) == 2
        ), "Two SELECT query should have been executed for test_table"


def test_file_based_database():
    """Test Duckmirror functionality with a file-based database."""
    # Create a temporary directory for our test database
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "test.db")

    try:
        # Create a new Duckmirror instance with file-based database
        mirror = Duckmirror(db_path, cache_in_memory=False)

        # Track function calls
        call_count = 0

        # Register a test function
        @mirror.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count
            call_count += 1
            return generate_test_df("mixed")

        # First access should create the table and return data
        df1 = mirror.df_test_table
        expected_df = generate_test_df("mixed")
        pd.testing.assert_frame_equal(df1, expected_df)
        assert call_count == 1, "Function should be called once on first access"

        # Verify the database file exists
        assert os.path.exists(db_path), "Database file should be created"

        # Create a new mirror instance pointing to the same database
        mirror2 = Duckmirror(db_path, cache_in_memory=False)

        # Track function calls for second instance
        call_count2 = 0

        # Register the same function
        @mirror2.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count2
            call_count2 += 1
            return generate_test_df("mixed")

        # Access should return the same data
        df2 = mirror2.df_test_table
        pd.testing.assert_frame_equal(df2, expected_df)
        assert (
            call_count2 == 1
        ), "Function should be called once on first access in second instance"

        # Second access should use cache
        df2_2 = mirror2.df_test_table
        pd.testing.assert_frame_equal(df2_2, expected_df)
        assert call_count2 == 1, "Function should not be called again on second access"

        # Modify the function to return different data
        call_count = 0  # Reset counter for modified function

        @mirror.register
        def test_table() -> pd.DataFrame:
            nonlocal call_count
            call_count += 1
            return generate_test_df("numbers")

        # Access should return the new data
        df3 = mirror.df_test_table
        new_expected_df = generate_test_df("numbers")
        pd.testing.assert_frame_equal(df3, new_expected_df)
        assert (
            call_count == 1
        ), "Modified function should be called once on first access"

        # Second access should use cache
        df3_2 = mirror.df_test_table
        pd.testing.assert_frame_equal(df3_2, new_expected_df)
        assert (
            call_count == 1
        ), "Modified function should not be called again on second access"

        # Verify we get copies of the dataframes
        assert df1 is not df2, "Dataframes should be different objects"
        assert df2 is not df3, "Dataframes should be different objects"
        assert df3 is not df3_2, "Dataframes should be different objects (copies)"

        # Verify the dataframes are not equal after modification
        pd.testing.assert_frame_equal(df1, df2)  # Should be equal (both mixed)
        with pytest.raises(AssertionError):
            pd.testing.assert_frame_equal(
                df1, df3
            )  # Should not be equal (mixed vs numbers)

    finally:
        # Clean up the temporary directory
        shutil.rmtree(temp_dir)


def test_overwrite_removes_existing_file():
    """Test that the overwrite parameter correctly handles existing database files."""
    # Create a temporary directory for our test
    temp_dir = tempfile.mkdtemp()
    db_path = os.path.join(temp_dir, "overwrite_test.db")

    try:
        # Phase 1: Initial database creation
        assert not os.path.exists(db_path), "Database file should not exist initially"

        mirror = Duckmirror(db_path, overwrite=False, cache_in_memory=False)

        @mirror.register
        def test_table() -> pd.DataFrame:
            return pd.DataFrame({"a": [1, 2, 3]})

        df1 = mirror.df_test_table
        pd.testing.assert_frame_equal(df1, pd.DataFrame({"a": [1, 2, 3]}))
        assert os.path.exists(
            db_path
        ), "Database file should exist after first creation"

        # Get initial file creation time
        initial_file_stats = os.stat(db_path)
        initial_creation_time = initial_file_stats.st_ctime

        # Close the connection to ensure file is released
        mirror.conn.close()

        # Phase 2: Overwrite with new data
        # File should still exist before overwrite
        assert os.path.exists(db_path), "Database file should exist before overwrite"

        # Create a new mirror with overwrite=True - this should delete the existing file
        mirror2 = Duckmirror(db_path, overwrite=True, cache_in_memory=False)

        # Verify the file was deleted and recreated by checking creation time
        assert os.path.exists(db_path), "Database file should exist after overwrite"
        new_file_stats = os.stat(db_path)
        new_creation_time = new_file_stats.st_ctime
        assert (
            new_creation_time > initial_creation_time
        ), "File should have been deleted and recreated with a newer timestamp"

        # Register and access the table to ensure it's using the new database
        @mirror2.register
        def test_table() -> pd.DataFrame:
            return pd.DataFrame({"a": [42, 43, 44]})

        df2 = mirror2.df_test_table
        pd.testing.assert_frame_equal(df2, pd.DataFrame({"a": [42, 43, 44]}))

        # Phase 3: Verify data was overwritten
        assert os.path.exists(
            db_path
        ), "Database file should still exist after overwrite"
        assert not df2.equals(
            pd.DataFrame({"a": [1, 2, 3]})
        ), "Old data should not be present after overwrite"

        # Close the second connection
        mirror2.conn.close()

    finally:
        # Clean up
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)


def test_class_methods_memory_duckmirror():
    """Test Duckmirror with a class (TestData) caching two different methods (df1, df2) in memory."""
    mirror = Duckmirror(":memory:", cache_in_memory=True)

    class TestData:
        @mirror.register
        def df1(self):
            return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

        @mirror.register
        def df2(self):
            return pd.DataFrame({"x": [10, 20, 30], "y": [40, 50, 60]})

    obj = TestData()

    # Access both cached DataFrames via the decorated class
    df1_result = obj.df1
    df2_result = obj.df2

    expected_df1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    expected_df2 = pd.DataFrame({"x": [10, 20, 30], "y": [40, 50, 60]})

    pd.testing.assert_frame_equal(df1_result, expected_df1)
    pd.testing.assert_frame_equal(df2_result, expected_df2)

    # Ensure cache returns copies, not the same object
    df1_result2 = obj.df1
    assert df1_result is not df1_result2
    pd.testing.assert_frame_equal(df1_result2, expected_df1)

    # Access both cached DataFrames via the mirror object
    mirror_df1 = mirror.TestData_df1
    mirror_df2 = mirror.TestData_df2

    pd.testing.assert_frame_equal(mirror_df1, expected_df1)
    pd.testing.assert_frame_equal(mirror_df2, expected_df2)

    # Ensure cache returns copies, not the same object
    mirror_df1_2 = mirror.TestData_df1
    assert mirror_df1 is not mirror_df1_2
    pd.testing.assert_frame_equal(mirror_df1_2, expected_df1)



import sys


class PrintCatcher(object):
    def __init__(self, stream="stdout"):
        self.text = ""
        if stream not in {"stdout", "stderr"}:
            raise ValueError('stream must be either "stdout" or "stderr"')
        self.stream = stream

    def write(self, text):
        self.text += text

    def flush(self):
        pass

    def __enter__(self):
        if self.stream == "stdout":
            sys.stdout = self
        else:
            sys.stderr = self
        return self

    def __exit__(self, *args):
        if self.stream == "stdout":
            sys.stdout = sys.__stdout__
        else:
            sys.stderr = sys.__stderr__
